# copyright 2004-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
from logilab.common.testlib import TestCase, unittest_main, mock_object

from yams.constraints import *
# after import *
from datetime import datetime, date, timedelta

class ConstraintTC(TestCase):

    def test_interval_serialization_integers(self):
        cstr = IntervalBoundConstraint(12, 13)
        self.assertEqual(cstr.serialize(), '12;13')
        cstr = IntervalBoundConstraint(maxvalue=13)
        self.assertEqual(cstr.serialize(), 'None;13')
        cstr = IntervalBoundConstraint(minvalue=13)
        self.assertEqual(cstr.serialize(), '13;None')
        self.assertRaises(AssertionError, IntervalBoundConstraint)

    def test_interval_serialization_floats(self):
        cstr = IntervalBoundConstraint(12.13, 13.14)
        self.assertEqual(cstr.serialize(), '12.13;13.14')


    def test_interval_deserialization_integers(self):
        cstr = IntervalBoundConstraint.deserialize('12;13')
        self.assertEqual(cstr.minvalue, 12)
        self.assertEqual(cstr.maxvalue, 13)
        cstr = IntervalBoundConstraint.deserialize('None;13')
        self.assertEqual(cstr.minvalue, None)
        self.assertEqual(cstr.maxvalue, 13)
        cstr = IntervalBoundConstraint.deserialize('12;None')
        self.assertEqual(cstr.minvalue, 12)
        self.assertEqual(cstr.maxvalue, None)

    def test_interval_deserialization_floats(self):
        cstr = IntervalBoundConstraint.deserialize('12.13;13.14')
        self.assertEqual(cstr.minvalue, 12.13)
        self.assertEqual(cstr.maxvalue, 13.14)


    def test_regexp_serialization(self):
        cstr = RegexpConstraint('[a-z]+,[A-Z]+', 12)
        self.assertEqual(cstr.serialize(), '[a-z]+,[A-Z]+,12')

    def test_regexp_deserialization(self):
        cstr = RegexpConstraint.deserialize('[a-z]+,[A-Z]+,12')
        self.assertEqual(cstr.regexp, '[a-z]+,[A-Z]+')
        self.assertEqual(cstr.flags, 12)

    def test_interval_with_attribute(self):
        cstr = IntervalBoundConstraint(NOW(), Attribute('hop'))
        cstr2 = IntervalBoundConstraint.deserialize(cstr.serialize())
        self.assertEqual(cstr2.minvalue.offset, None)
        self.assertEqual(cstr2.maxvalue.attr, 'hop')
        self.assertTrue(cstr2.check(mock_object(hop=datetime.now()+timedelta(hours=1)),
                                    'hip', datetime.now() + timedelta(seconds=2)))
        # fail, value < minvalue
        self.assertFalse(cstr2.check(mock_object(hop=datetime.now()+timedelta(hours=1)),
                                'hip', datetime.now() - timedelta(hours=2)))
        # fail, value > maxvalue
        self.assertFalse(cstr2.check(mock_object(hop=datetime.now()+timedelta(hours=1)),
                                'hip', datetime.now() + timedelta(hours=2)))


    def test_interval_with_date(self):
        cstr = IntervalBoundConstraint(TODAY(timedelta(1)),
                                       TODAY(timedelta(3)))
        cstr2 = IntervalBoundConstraint.deserialize(cstr.serialize())
        self.assertEqual(cstr2.minvalue.offset, timedelta(1))
        self.assertEqual(cstr2.maxvalue.offset, timedelta(3))
        self.assertTrue(cstr2.check(None, 'hip', date.today() + timedelta(2)))
        # fail, value < minvalue
        self.assertFalse(cstr2.check(None, 'hip', date.today()))
        # fail, value > maxvalue
        self.assertFalse(cstr2.check(None, 'hip', date.today() + timedelta(4)))

    def test_bound_with_attribute(self):
        cstr = BoundaryConstraint('<=', Attribute('hop'))
        cstr2 = BoundaryConstraint.deserialize(cstr.serialize())
        self.assertEqual(cstr2.boundary.attr, 'hop')
        self.assertEqual(cstr2.operator, '<=')
        self.assertTrue(cstr2.check(mock_object(hop=date.today()), 'hip', date.today()))
        # fail, value > maxvalue
        self.assertFalse(cstr2.check(mock_object(hop=date.today()),
                                'hip', date.today() + timedelta(days=1)))


    def test_bound_with_date(self):
        cstr = BoundaryConstraint('<=', TODAY())
        cstr2 = BoundaryConstraint.deserialize(cstr.serialize())
        self.assertEqual(cstr2.boundary.offset, None)
        self.assertEqual(cstr2.operator, '<=')
        self.assertTrue(cstr2.check(None, 'hip', date.today()))
        # fail, value > maxvalue
        self.assertFalse(cstr2.check(None, 'hip', date.today() + timedelta(days=1)))

    def test_vocab_constraint_serialization(self):
        cstr = StaticVocabularyConstraint(['a, b', 'c'])
        self.assertEqual(StaticVocabularyConstraint.deserialize(cstr.serialize()).values,
                         ('a, b', 'c'))

if __name__ == '__main__':
    unittest_main()





# -*- coding: iso-8859-1 -*-
# copyright 2004-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""unit tests for module yams.schema classes"""

from __future__ import with_statement

from logilab.common.testlib import TestCase, unittest_main

from copy import copy, deepcopy
from tempfile import mktemp

from yams import (BASE_TYPES, ValidationError, BadSchemaDefinition,
                  register_base_type, unregister_base_type)
from yams.buildobjs import (register_base_types, make_type, _add_relation,
                            EntityType, RelationType, RelationDefinition)
from yams.schema import Schema, RelationDefinitionSchema
from yams.interfaces import IVocabularyConstraint
from yams.constraints import (BASE_CHECKERS, SizeConstraint, RegexpConstraint,
                              StaticVocabularyConstraint, IntervalBoundConstraint,
                              FormatConstraint)
from yams.reader import SchemaLoader


# build a dummy schema ########################################################


class BaseSchemaTC(TestCase):
    def setUp(self):
        global schema, enote, eaffaire, eperson, esociete, estring, eint
        global rconcerne, rnom
        schema = Schema('Test Schema')
        register_base_types(schema)
        enote = schema.add_entity_type(EntityType('Note'))
        eaffaire = schema.add_entity_type(EntityType('Affaire'))
        eperson = schema.add_entity_type(EntityType('Person'))
        esociete = schema.add_entity_type(EntityType('Societe'))


        RELS = (
            # attribute relations
            ('Note date Datetime'),
            ('Note type String'),
            ('Affaire sujet String'),
            ('Affaire ref String'),
            ('Affaire starton Time'),
            ('Person nom String'),
            ('Person prenom String'),
            ('Person sexe Float'),
            ('Person tel Int'),
            ('Person fax Int'),
            ('Person datenaiss Date'),
            ('Person TEST Boolean'),
            ('Person promo String'),
            ('Person promo_enlarged String'),
            ('Person promo_encoding String'),
            ('Person promo_format String'),
            # real relations
            ('Person  travaille Societe'),
            ('Person  evaluee   Note'),
            ('Societe evaluee   Note'),
            ('Person  concerne  Affaire'),
            ('Person  concerne  Societe'),
            ('Affaire concerne  Societe'),
            )
        for i, rel in enumerate(RELS):
            _from, _type, _to = rel.split()
            try:
                schema.rschema(_type)
            except KeyError:
                schema.add_relation_type(RelationType(_type))
            schema.add_relation_def(RelationDefinition(_from, _type, _to, order=i))
        schema.rschema('nom').rdef('Person', 'String').cardinality = '11' # not null

        enote.rdef('type').constraints = [StaticVocabularyConstraint((u'bon', u'pasbon',
                                                                      u'bof', u'peux mieux faire'))]
        enote.rdef('date').cardinality = '11'

        eaffaire.rdef('sujet').constraints = [SizeConstraint(128)]
        eaffaire.rdef('ref').constraints = [SizeConstraint(12), RegexpConstraint(r'[A-Z]+\d+')]
        eperson.rdef('nom').constraints = [SizeConstraint(20, 10)]
        eperson.rdef('prenom').constraints = [SizeConstraint(64)]
        eperson.rdef('tel').constraints = [IntervalBoundConstraint(maxvalue=999999)]
        eperson.rdef('fax').constraints = [IntervalBoundConstraint(minvalue=12, maxvalue=999999)]
        eperson.rdef('promo').constraints = [StaticVocabularyConstraint( (u'bon', u'pasbon'))]
        eperson.rdef('promo_format').constraints = [FormatConstraint()]

        estring = schema.eschema('String')
        eint = schema.eschema('Int')
        rconcerne = schema.rschema('concerne')
        rnom = schema.rschema('nom')

    def assertRaisesMsg(self, ex_class, msg, func, *args, **kwargs):
        self.assertRaises(ex_class, func, *args, **kwargs)
        try:
            func(*args, **kwargs)
        except Exception, ex:
            self.assertEqual(str(ex), msg)

# test data ###################################################################

BAD_RELS = ( ('Truc badrelation1 Affaire'),
             ('Affaire badrelation2 Machin'),
             )

ATTRIBUTE_BAD_VALUES = (
    ('Person', [('nom', 1), ('nom', u'tropcour'),
                ('nom', u'>10 mais  supérieur à < 20 , c\'est long'),
                ('sexe', u'F'), ('sexe', u'MorF'), ('sexe', 'F'),
                ('promo', 'bon'), ('promo', 'uyou'),
                ('promo', u' pas bon du tout'),
                ('promo_format', u'text/something'),
                ('tel', 'notastring'),
                ('tel', 1000000),
                ('fax', 11),
                ('TEST', 'notaboolean'), #('TEST', 0), ('TEST', 1)]), #should we accept this ?
                ('TEST', 'true'), ('TEST', 'false')]),
## the date and time are not checked for now
##    ('Person', [('nom', u' >10 mais < 20 '),
##               ('datenaiss', '979-06-12')]),
##    ('Note', [('date', '2229-01-31 minuit')]),
##    ('Affaire', [('starton', 'midi')]),

    ('Note', [('type', ['bof', 'peux mieux faire']),
              ('type', 'bof, je suis pas unicode, alors...'),
              ('date', None),
              ]),
    ('Affaire', [('ref', 'ginco01'), ('ref', 'GINCO'),],
    ),
    )

ATTRIBUTE_GOOD_VALUES = (
    ('Person', [('nom', u'>10 mais < 20 '), ('sexe', 0.5),
                ('promo', u'bon'),
                ('datenaiss', '1977-06-07'),
                ('tel', 83433), ('fax', None), ('fax', 12),
                ('TEST', True), ('TEST', False)]),
    ('Note', [('date', '2229-01-31 00:00')]),
    ('Affaire', [('starton', '00:00'),
                 ('ref', u'GINCO01')]),
    )

RELATIONS_BAD_VALUES = {
    'travaille': [('Person', 'Affaire'), ('Affaire', 'Societe'),
                  ('Affaire', 'Societe'), ('Societe', 'Person')]
    }
RELATIONS_GOOD_VALUES = {
    'travaille': [('Person', 'Societe')],
    'concerne': [('Person', 'Affaire'), ('Affaire', 'Societe')]
    }


# test suite ##################################################################

class EntitySchemaTC(BaseSchemaTC):

    def test_base(self):
        self.assert_(repr(eperson))

    def test_cmp(self):
        self.assertTrue(eperson == 'Person')
        self.assertTrue('Person' == eperson)
        self.assertTrue(eperson != 'Note')
        self.assertTrue('Note' != eperson)
        self.assertFalse(enote == eperson)
        self.assertFalse(eperson == enote)
        self.assertTrue(enote != eperson)
        self.assertTrue(eperson != enote)
        l = [eperson, enote, eaffaire, esociete]
        l.sort()
        self.assertListEqual(l, [eaffaire, enote, eperson, esociete])
        self.assertListEqual(l, ['Affaire', 'Note', 'Person', 'Societe'])

    def test_hash(self):
        d = {}
        d[eperson] = 'p'
        d[enote] = 'n'
        self.assertEqual(d[copy(eperson)], 'p')
        self.assertEqual(d[copy(enote)], 'n')
        self.assertEqual(d['Person'], 'p')
        self.assertEqual(d['Note'], 'n')
        d = {}
        d['Person'] = eperson
        d['Note'] = enote
        self.assertEqual(copy(eperson), 'Person')
        self.assertEqual(d[copy(eperson)], 'Person')
        self.assertEqual(d[copy(enote)], 'Note')

    def test_deepcopy_with_regexp_constraints(self):
        eaffaire.rdef('ref').constraints = [RegexpConstraint(r'[A-Z]+\d+')]
        rgx_cstr, = eaffaire.rdef('ref').constraints
        eaffaire2 = deepcopy(schema).eschema('Affaire')
        rgx_cstr2, = eaffaire2.rdef('ref').constraints
        self.assertEqual(rgx_cstr2.regexp, rgx_cstr.regexp)
        self.assertEqual(rgx_cstr2.flags, rgx_cstr.flags)
        self.assertEqual(rgx_cstr2._rgx, rgx_cstr._rgx)

    def test_deepcopy(self):
        global schema
        schema = deepcopy(schema)
        self.assertFalse(eperson is schema['Person'])
        self.assertEqual(eperson, schema['Person'])
        self.assertEqual('Person', schema['Person'])
        self.assertEqual(eperson.subject_relations(), schema['Person'].subject_relations())
        self.assertEqual(eperson.object_relations(), schema['Person'].object_relations())
        self.assertEqual(schema.eschema('Person').final, False)
        self.assertEqual(schema.eschema('String').final, True)
        self.assertEqual(schema.rschema('ref').final, True)
        self.assertEqual(schema.rschema('concerne').final, False)

    def test_attribute_description(self):
        schema = SchemaLoader().load([self.datadir], 'Test')
        self.assertEqual(schema['EPermission'].rdef('name').description,
                         'name or identifier of the permission')

    def test_deepcopy_specialization(self):
        schema2 = deepcopy(SchemaLoader().load([self.datadir], 'Test'))
        edivision = schema2.eschema('Division')
        self.assertEqual(edivision.specializes(), 'Company')
        self.assertEqual(edivision.specialized_by(), ['Subdivision'])
        schema2.del_entity_type('Subdivision')
        self.assertEqual(edivision.specialized_by(), [])

    def test_is_final(self):
        self.assertEqual(eperson.final, False)
        self.assertEqual(enote.final, False)
        self.assertEqual(estring.final, True)
        self.assertEqual(eint.final, True)
        self.assertEqual(eperson.subjrels['nom'].final, True)
        #self.assertEqual(eperson.is_final('concerne'), False)
        self.assertEqual(eperson.subjrels['concerne'].final, False)

    def test_is_metadata(self):
        self.assertEqual(eperson.is_metadata('promo'), None)
        self.assertEqual(eperson.is_metadata('promo_enlarged'), None)
        self.assertEqual(eperson.is_metadata('promo_encoding'), ('promo', 'encoding'))
        self.assertEqual([(k.type, v)  for k, v in eperson.meta_attributes().items()],
                          [('promo_encoding', ('encoding', 'promo')),
                           ('promo_format', ('format', 'promo'))])

    def test_defaults(self):
        self.assertEqual(list(eperson.defaults()), [])
        self.assertRaises(StopIteration, estring.defaults().next)

    def test_vocabulary(self):
        #self.assertEqual(eperson.vocabulary('promo')
        self.assertEqual(eperson.rdef('promo').constraint_by_interface(IVocabularyConstraint).vocabulary(),
                          ('bon', 'pasbon'))
        # self.assertRaises(AssertionError,
        #                   eperson.vocabulary, 'nom')

    def test_indexable_attributes(self):
        eperson.rdef('nom').fulltextindexed = True
        eperson.rdef('prenom').fulltextindexed = True
        self.assertEqual(list(eperson.indexable_attributes()), ['nom', 'prenom'])


    def test_goodValues_relation_default(self):
        """check good values of entity does not raise an exception"""
        eperson.rdef('nom').default = 'No name'
        self.assertEqual(eperson.default('nom'), 'No name')

    def test_subject_relations(self):
        """check subject relations a returned in the same order as in the
        schema definition"""
        rels = eperson.ordered_relations()
        expected = ['nom', 'prenom', 'sexe', 'tel', 'fax', 'datenaiss', 'TEST',
                    'promo', 'promo_enlarged', 'promo_encoding', 'promo_format',
                    'travaille', 'evaluee', 'concerne']
        self.assertEqual([r.type for r in rels], expected)

    def test_object_relations(self):
        """check object relations a returned in the same order as in the
        schema definition"""
        rels = eaffaire.object_relations()
        expected = ['concerne']
        self.assertEqual(rels, expected)
        rels = [schem.type for schem in eaffaire.object_relations()]
        self.assertEqual(rels, expected)
        self.assertEqual(eaffaire.objrels['concerne'].type,
                         'concerne')

    def test_destination_type(self):
        """check subject relations a returned in the same order as in the
        schema definition"""
        self.assertEqual(eperson.destination('nom'), 'String')
        self.assertEqual(eperson.destination('travaille'), 'Societe')

    def test_check_unique_together1(self):
        eperson._unique_together = [('prenom', 'nom')]
        eperson.check_unique_together()

    def test_check_unique_together2(self):
        eperson._unique_together = [('prenom', 'noms')]
        with self.assertRaises(BadSchemaDefinition) as cm:
            eperson.check_unique_together()
        self.assertTrue('no such attribute or relation noms'
                        in cm.exception.args[0])

    def test_check_unique_together3(self):
        eperson._unique_together = [('nom', 'travaille')]
        with self.assertRaises(BadSchemaDefinition) as cm:
            eperson.check_unique_together()
        self.assertTrue('travaille is not an attribute or an inlined relation'
                        in cm.exception.args[0])


class RelationSchemaTC(BaseSchemaTC):

    def test_cmp(self):
        self.assertTrue(rconcerne == 'concerne')
        self.assertTrue('concerne' == rconcerne)
        self.assertTrue(rconcerne != 'nom')
        self.assertTrue('nom' != rconcerne)
        self.assertFalse(rnom == rconcerne)
        self.assertFalse(rconcerne == rnom)
        self.assertTrue(rnom != rconcerne)
        self.assertTrue(rconcerne != rnom)

    def test_hash(self):
        d = {}
        d[rconcerne] = 'p'
        d[rnom] = 'n'
        self.assertEqual(d[copy(rconcerne)], 'p')
        self.assertEqual(d[copy(rnom)], 'n')
        self.assertEqual(d['concerne'], 'p')
        self.assertEqual(d['nom'], 'n')


    def test_base(self):
        self.assert_(repr(rnom))

    def test_star_types(self):
        types = sorted(rconcerne.subjects())
        self.assertEqual(types, ['Affaire', 'Person'])
        types = sorted(rconcerne.objects())
        self.assertEqual(types, ['Affaire', 'Societe'])

    def test_raise_update(self):
        self.assertRaisesMsg(BadSchemaDefinition,
                             'type String can\'t be used as subject in a relation',
                             rconcerne.update, estring, enote, {})
##         self.assertRaisesMsg(BadSchemaDefinition,
##                              "can't have a final relation pointing to multiple entity types (nom: ['String', 'Int'])" ,
##                              rnom.update, enote, eint)
        self.assertRaisesMsg(BadSchemaDefinition, 'ambiguous relation nom: String is final but not Affaire',
                             rnom.update, enote, eaffaire, {})
        self.assertRaises(BadSchemaDefinition,
                          rconcerne.update, enote, estring, {})

    def test_association_types(self):
        expected = [ ('Affaire', ['Societe']),
                     ('Person', ['Affaire', 'Societe']) ]
        assoc_types = rconcerne.associations()
        assoc_types.sort()
        self.assertEqual(assoc_types, expected)
        assoc_types = []
        for _from, _to in rconcerne.associations():
            assoc_types.append( (_from, _to))
            #assoc_types.append( (_from.type, [s.type for s in _to]) )
        assoc_types.sort()
        self.assertEqual(assoc_types, expected)

#     def test_reverse_association_types(self):
#         expected = [ ('Affaire', ['Person']),
#                      ('Societe', ['Person', 'Affaire'])]
#         assoc_types = rconcerne.reverse_association_types()
#         assoc_types.sort()
#         self.assertEqual(assoc_types, expected)
#         assoc_types = []
#         for _from, _to in rconcerne.reverse_association_types(True):
#             assoc_types.append( (_from.type, [s.type for s in _to]) )
#         assoc_types.sort()
#         self.assertEqual(assoc_types, expected)


class SchemaTC(BaseSchemaTC):

    def test_schema_base(self):
        """test base schema methods
        """
        all_types = ['Affaire', 'BigInt', 'Boolean', 'Bytes', 'Date', 'Datetime',
                     'Decimal',
                     'Float', 'Int', 'Interval', 'Note', 'Password',
                     'Person', 'Societe', 'String', 'TZDatetime', 'TZTime', 'Time']
        self.assertEqual(sorted(schema.entities()), all_types)
        self.assertEqual(schema.has_entity('Affaire'), True)
        self.assertEqual(schema.has_entity('Aaire'), False)

    def test_raise_add_entity_type(self):
        self.assertRaisesMsg(BadSchemaDefinition, "entity type Person is already defined" ,
                             schema.add_entity_type, EntityType('Person'))

    def test_raise_relation_def(self):
        self.assertRaisesMsg(BadSchemaDefinition, "using unknown type 'Afire' in relation evaluee"  ,
                             schema.add_relation_def, RelationDefinition('Afire', 'evaluee', 'Note'))
## XXX what is this ?
##        self.assertRaisesMsg(BadSchemaDefinition, 'the "symmetric" property should appear on every definition of relation evaluee' ,
##                             schema.add_relation_def, RelationDefinition('Affaire', 'evaluee', 'Note', symmetric=True))

    def test_schema_relations(self):
        all_relations = ['TEST', 'concerne', 'travaille', 'evaluee',
                         'date', 'type', 'sujet', 'ref', 'nom', 'prenom',
                         'starton', 'sexe', 'promo', 'promo_enlarged',
                         'promo_encoding', 'promo_format', 'tel', 'fax',
                         'datenaiss']
        all_relations.sort()
        relations = schema.relations()
        relations.sort()
        self.assertEqual(relations, all_relations)

        self.assertEqual(len(eperson.rdef('nom').constraints), 1)
        self.assertEqual(len(eperson.rdef('prenom').constraints), 1)

    def test_schema_check_relations(self):
        """test behaviour with some incorrect relations"""
        for rel in BAD_RELS:
            _from, _type, _to = rel.split()
            self.assertRaises(BadSchemaDefinition,
                              schema.add_relation_def, RelationDefinition(_from, _type, _to))
        # check we can't extend a final relation
        self.assertRaises(BadSchemaDefinition,
                          schema.add_relation_def, RelationDefinition('Person', 'nom', 'affaire'))

    def test_entities_goodValues_check(self):
        """check good values of entity does not raise an exception"""
        for etype, val_list in ATTRIBUTE_GOOD_VALUES:
            eschema = schema.eschema(etype)
            eschema.check(dict(val_list))

    def test_entities_badValues_check(self):
        """check bad values of entity raises ValidationError exception"""
        for etype, val_list in ATTRIBUTE_BAD_VALUES:
            eschema = schema.eschema(etype)
            # check attribute values one each time...
            for item in val_list:
                with self.assertRaises(ValidationError) as cm:
                    eschema.check(dict([item]))
                # check automatic call to translation works properly
                unicode(cm.exception)

    def test_validation_error_translation_1(self):
        eschema = schema.eschema('Person')
        with self.assertRaises(ValidationError) as cm:
            eschema.check({'nom': 1, 'promo': 2})
        cm.exception.translate(unicode)
        self.assertEqual(cm.exception.errors,
                         {'nom-subject': u'incorrect value (1) for type "String"',
                          'promo-subject': u'incorrect value (2) for type "String"'})

    def test_validation_error_translation_2(self):
        eschema = schema.eschema('Person')
        with self.assertRaises(ValidationError) as cm:
            eschema.check({'nom': u'x'*21, 'prenom': u'x'*65})
        cm.exception.translate(unicode)
        self.assertEqual(cm.exception.errors,
                         {'nom-subject': u'value should have maximum size of 20 but found 21',
                          'prenom-subject': u'value should have maximum size of 64 but found 65'})

    def test_validation_error_translation_3(self):
        eschema = schema.eschema('Person')
        with self.assertRaises(ValidationError) as cm:
            eschema.check({'tel': 1000000, 'fax': 1000001})
        cm.exception.translate(unicode)
        self.assertEqual(cm.exception.errors,
                         {'fax-subject': u'value 1000001 must be <= 999999',
                          'tel-subject': u'value 1000000 must be <= 999999'})

    def test_validation_error_translation_4(self):
        verr = ValidationError(1, {None: 'global message about eid %(eid)s'}, {'eid': 1})
        verr.translate(unicode)
        self.assertEqual(verr.errors,
                         {None: 'global message about eid 1'})

    def test_pickle(self):
        """schema should be pickeable"""
        import pickle
        picklefile = mktemp()
        picklestream = open(picklefile, 'w')
        pickle.dump(schema, picklestream)
        picklestream.close()
        pschema = pickle.load(open(picklefile))
        self.assertFalse(eperson is pschema['Person'])
        self.assertEqual(eperson, pschema['Person'])
        self.assertEqual('Person', pschema['Person'])
        self.assertEqual(eperson.ordered_relations(), pschema['Person'].ordered_relations())
        self.assertEqual(eperson.object_relations(), pschema['Person'].object_relations())


    def test_rename_entity_type(self):
        affaire = schema.eschema('Affaire')
        orig_rprops = affaire.rdef('concerne')
        schema.rename_entity_type('Affaire', 'Workcase')
        self.assertItemsEqual(schema._entities.keys(),
                             ['BigInt', 'Boolean', 'Bytes', 'Date', 'Datetime', 'Float',
                              'Decimal',
                              'Int', 'Interval', 'Note', 'Password', 'Person',
                              'Societe', 'String', 'Time', 'TZDatetime', 'TZTime',
                              'Workcase'])
        rconcerne = schema.rschema('concerne')
        self.assertItemsEqual(rconcerne.subjects(), ['Workcase', 'Person'])
        self.assertItemsEqual(rconcerne.objects(), ['Workcase', 'Societe'])
        self.assertRaises(KeyError, schema.eschema, 'Affaire')
        workcase = schema.eschema('Workcase')
        schema.__test__ = True
        self.assertEqual(workcase.rdef('concerne'), orig_rprops)

    def test_inheritance_rdefs(self):
        class Plan(EntityType):
            pass
        rdef = RelationDefinition('Plan', 'custom_workflow', 'Workflow')
        _add_relation(Plan.__relations__, rdef)
        class TE(Plan):
            pass
        self.assertListEqual(['custom_workflow'],
                             [rel.name for rel in TE.__relations__])


class SymetricTC(TestCase):
    def setUp(self):
        global schema
        schema = Schema('Test Schema')
        schema.add_entity_type(EntityType('Bug'))
        schema.add_entity_type(EntityType('Story'))
        schema.add_entity_type(EntityType('Project'))
        schema.add_relation_type(RelationType('see_also', symmetric=True))

    def test_association_types(self):
        schema.add_relation_def(RelationDefinition('Bug', 'see_also', 'Bug'))
        schema.add_relation_def(RelationDefinition('Bug', 'see_also', 'Story'))
        schema.add_relation_def(RelationDefinition('Bug', 'see_also', 'Project'))
        schema.add_relation_def(RelationDefinition('Story', 'see_also', 'Story'))
        schema.add_relation_def(RelationDefinition('Story', 'see_also', 'Project'))
        schema.add_relation_def(RelationDefinition('Project', 'see_also', 'Project'))

        rsee_also = schema.rschema('see_also')
        subj_types = rsee_also.associations()
        subj_types.sort()
        self.assertEqual(subj_types,
                          [('Bug', ['Bug', 'Story', 'Project']),
                           ('Project', ['Bug', 'Story', 'Project']),
                           ('Story', ['Bug', 'Story', 'Project'])])

    def test_wildcard_association_types(self):
        class see_also(RelationDefinition):
            subject = '*'
            object ='*'
        see_also.expand_relation_definitions({'see_also': see_also}, schema)
        rsee_also = schema.rschema('see_also')
        subj_types = rsee_also.associations()
        subj_types.sort()
        for key, vals in subj_types:
            vals.sort()
        self.assertEqual(subj_types,
                          [('Bug', ['Bug', 'Project', 'Story']),
                           ('Project', ['Bug', 'Project', 'Story']),
                           ('Story', ['Bug', 'Project', 'Story'])])


class CustomTypeTC(TestCase):

    def tearDown(self):
        try:
            unregister_base_type('Test')
        except AssertionError:
            pass

    def test_register_base_type(self):
        register_base_type('Test', ('test1', 'test2'))
        self.assertIn('Test', BASE_TYPES)
        self.assertIn('Test', RelationDefinitionSchema.BASE_TYPE_PROPERTIES)
        self.assertEqual(RelationDefinitionSchema.BASE_TYPE_PROPERTIES['Test'],
                         {'test1': None, 'test2': None})
        self.assertTrue('Test' in BASE_CHECKERS)

    def test_make_base_type_class(self):
        register_base_type('Test', ('test1', 'test2'))
        Test = make_type('Test')
        self.assertIsInstance(Test, type)
        self.assertEqual(Test.etype, 'Test')
        t = Test(test1=1)
        self.assertEqual(t.test1, 1)
        self.assertFalse(hasattr(t, 'test2'))

if __name__ == '__main__':
    unittest_main()



# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""unit tests for module yams.diff"""
import os.path as osp

from logilab.common.testlib import TestCase, unittest_main

from yams.reader import build_schema_from_namespace
from yams.buildobjs import String, Int, Float, EntityType, SubjectRelation
from yams.diff import properties_from, schema_diff


class PersonBase(EntityType):
    nom    = String()
    prenom = String()

class PersonAttrMod(EntityType):
    nom    = String()
    prenom = Float()

class PersonAttrAdd(EntityType):
    nom    = String()
    prenom = String()
    age = Int()
    is_friend_of = SubjectRelation('PersonAttrMod')

class PersonAttrAdd2(EntityType):
    nom    = String()
    prenom = String()
    salaire = Float()
    is_friend_of = SubjectRelation('PersonAttrAdd')

class PersonAttrAdd3(EntityType):
    nom     = String()
    prenom  = String()
    age     = Int()
    salaire = Float()
    is_spouse_of = SubjectRelation('PersonAttrAdd')

def create_schema_1():
    class Affaire(EntityType):
        nom = String()
        associate_affaire = SubjectRelation('Affaire')
    return build_schema_from_namespace([('PersonBase', PersonBase),
                                        ('Affaire',    Affaire),
                                        ('PersonAttrMod', PersonAttrMod)])


def create_schema_2():
    class Affaire(EntityType):
        nom = String(maxsize=150)
        numero = Int()
        associate_affaire = SubjectRelation('Affaire', cardinality='**')
        associate_person = SubjectRelation('PersonBase')
    return build_schema_from_namespace([('PersonBase', PersonBase),
                                        ('Affaire',    Affaire)])

class PropertiesFromTC(TestCase):

    def build_rdef(self, props_ref):
        class AType(EntityType):
            attr = String(**props_ref)
        schema = build_schema_from_namespace(locals().items())
        return schema['AType'].rdef('attr')

    def build_props_dict(self, props_ref):
        result = {'default': None,
                  'description': '',
                  'fulltextindexed': False,
                  'indexed': False,
                  'internationalizable': False,
                  'order': 1,
                  'required': False,
                  'uid': False}
        result.update(props_ref)
        return result

    def test_properties_from_final_attributes_1(self):
        props_ref = {'required': True, 'default': 'toto',
                     'description': 'something'}
        self.assertEqual({'default': 'toto',
                          'required': True,
                          '__permissions__': " {\n\t\t\t'read': ('managers','users','guests'),\n\t\t\t'add': ('managers','users'),\n\t\t\t'update': ('managers','owners')\n\t\t}",
                          'description': "'something'",
                          'order': 1},
                         properties_from(self.build_rdef(props_ref)),
                         )

    def test_properties_from_final_attributes_2(self):
        props_ref = {}
        self.assertEqual({'__permissions__': " {\n\t\t\t'read': ('managers','users','guests'),\n\t\t\t'add': ('managers','users'),\n\t\t\t'update': ('managers','owners')\n\t\t}",
                          'order': 1},
                         properties_from(self.build_rdef(props_ref)))

    def test_properties_from_final_attributes_3(self):
        props_ref = {'default': None, 'required': False}
        self.assertEqual({'__permissions__': " {\n\t\t\t'read': ('managers','users','guests'),\n\t\t\t'add': ('managers','users'),\n\t\t\t'update': ('managers','owners')\n\t\t}",
                          'order': 1},
                         properties_from(self.build_rdef(props_ref)))

    def test_constraint_properties_1(self):
        props_ref = {'maxsize': 150, 'required': False}
        self.assertEqual({'maxsize': 150,
                          '__permissions__': " {\n\t\t\t'read': ('managers','users','guests'),\n\t\t\t'add': ('managers','users'),\n\t\t\t'update': ('managers','owners')\n\t\t}",
                          'order': 1},
                         properties_from(self.build_rdef(props_ref)))

    def test_constraint_properties_2(self):
        props_ref = {'unique': True, 'required': False}
        self.assertEqual({'__permissions__': " {\n\t\t\t'read': ('managers','users','guests'),\n\t\t\t'add': ('managers','users'),\n\t\t\t'update': ('managers','owners')\n\t\t}",
                          'unique': True,
                          'order': 1},
                         properties_from(self.build_rdef(props_ref)))

    def test_constraint_properties_3(self):
        props_ref = {'vocabulary': ('aaa', 'bbbb', 'ccccc'), 'required': False, 'maxsize': 20}
        rdef = self.build_rdef(props_ref)
        props_ref['maxsize'] = 5
        self.assertEqual({'maxsize': 5,
                          '__permissions__': " {\n\t\t\t'read': ('managers','users','guests'),\n\t\t\t'add': ('managers','users'),\n\t\t\t'update': ('managers','owners')\n\t\t}",
                          'order': 1,
                          'vocabulary': [u'aaa', u'bbbb', u'ccccc']},
                         properties_from(rdef))


class SchemaDiff(TestCase):

    def test_schema_diff(self):
        schema1 = create_schema_1()
        schema2 = create_schema_2()
        output1, output2 = schema_diff(schema1, schema2)
        self.assertMultiLineEqual(open(output1).read(),
                                  open(self.datapath('schema1.txt')).read())
        self.assertMultiLineEqual(open(output2).read(),
                                  open(self.datapath('schema2.txt')).read())


if __name__ == '__main__':
    unittest_main()



from logilab.common import testlib

from yams.buildobjs import register_base_types, String, EntityType, RelationType, RelationDefinition
from yams import schema, serialize

class MyTests(testlib.TestCase):

    def test_yams_serialize(self):
        s = schema.Schema('Unknown')
        register_base_types(s)
        e2 = EntityType('Ent2')
        s.add_entity_type(e2)
        e1 = EntityType('Ent1')
        e1.specialized_type = 'Ent2'
        s.add_entity_type(e1)
        s.add_relation_type(RelationType('attr1'))
        s.add_relation_def(RelationDefinition('Ent1', 'attr1', 'String'))
        out = serialize.serialize_to_python(s)
        self.assertMultiLineEqual(out, '\n'.join([
                'from yams.buildobjs import *',
                '',
                'class Ent2(EntityType):',
                '    pass',
                '',
                'class Ent1(Ent2):',
                '    attr1 = String()',
                '\n']))

if __name__ == '__main__':
    testlib.unittest_main()



# copyright 2004-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Some common constraint classes."""

__docformat__ = "restructuredtext en"
_ = unicode

import re
import decimal
import operator
from StringIO import StringIO

from logilab.common.deprecation import class_renamed

import yams
from yams import BadSchemaDefinition
from yams.interfaces import IConstraint, IVocabularyConstraint


class BaseConstraint(object):
    """base class for constraints"""
    __implements__ = IConstraint

    def check_consistency(self, subjschema, objschema, rdef):
        pass

    def type(self):
        return self.__class__.__name__

    def serialize(self):
        """called to make persistent valuable data of a constraint"""
        return None

    @classmethod
    def deserialize(cls, value):
        """called to restore serialized data of a constraint. Should return
        a `cls` instance
        """
        return cls()

    def failed_message(self, key, value):
        return _('%(KEY-cstr)s constraint failed for value %(KEY-value)r'), {
            key+'-cstr': self,
            key+'-value': value}

# possible constraints ########################################################

class UniqueConstraint(BaseConstraint):
    """object of relation must be unique"""

    def __str__(self):
        return 'unique'

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("unique constraint doesn't apply to non "
                                      "final entity type")

    def check(self, entity, rtype, values):
        """return true if the value satisfy the constraint, else false"""
        return True


class SizeConstraint(BaseConstraint):
    """the string size constraint :

    if max is not None the string length must not be greater than max
    if min is not None the string length must not be shorter than min
    """

    def __init__(self, max=None, min=None):
        assert (max is not None or min is not None), "No max or min"
        if min is not None:
            assert isinstance(min, int), 'min must be an int, not %r' % min
        if max is not None:
            assert isinstance(max, int), 'max must be an int, not %r' % max
        self.max = max
        self.min = min

    def __str__(self):
        res = 'size'
        if self.max is not None:
            res = '%s <= %s' % (res, self.max)
        if self.min is not None:
            res = '%s <= %s' % (self.min, res)
        return res

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("size constraint doesn't apply to non "
                                      "final entity type")
        if not objschema in ('String', 'Bytes', 'Password'):
            raise BadSchemaDefinition("size constraint doesn't apply to %s "
                                      "entity type" % objschema)
        if self.max:
            for cstr in rdef.constraints:
                if cstr.__class__ is StaticVocabularyConstraint:
                    for value in cstr.values:
                        if len(value) > self.max:
                            raise BadSchemaDefinition(
                                'size constraint set to %s but vocabulary '
                                'contains string of greater size' % self.max)

    def check(self, entity, rtype, value):
        """return true if the value is in the interval specified by
        self.min and self.max
        """
        if self.max is not None:
            if len(value) > self.max:
                return False
        if self.min is not None:
            if len(value) < self.min:
                return False
        return True

    def failed_message(self, key, value):
        if self.max is not None and len(value) > self.max:
            return _('value should have maximum size of %(KEY-max)s but found %(KEY-size)s'), {
                key+'-max': self.max,
                key+'-size': len(value)}
        if self.min is not None and len(value) < self.min:
            return _('value should have minimum size of %(KEY-min)s but found %(KEY-size)s'), {
                key+'-min': self.min,
                key+'-size': len(value)}
        assert False, 'shouldnt be there'

    def serialize(self):
        """simple text serialization"""
        if self.max and self.min:
            return u'min=%s,max=%s' % (self.min, self.max)
        if self.max:
            return u'max=%s' % (self.max)
        return u'min=%s' % (self.min)

    @classmethod
    def deserialize(cls, value):
        """simple text deserialization"""
        kwargs = {}
        for adef in value.split(','):
            key, val = [w.strip() for w in adef.split('=')]
            assert key in ('min', 'max')
            kwargs[str(key)] = int(val)
        return cls(**kwargs)


class RegexpConstraint(BaseConstraint):
    """specifies a set of allowed patterns for a string value"""
    __implements__ = IConstraint

    def __init__(self, regexp, flags=0):
        """
        Construct a new RegexpConstraint.

        :Parameters:
         - `regexp`: (str) regular expression that strings must match
         - `flags`: (int) flags that are passed to re.compile()
        """
        self.regexp = regexp
        self.flags = flags
        self._rgx = re.compile(regexp, flags)

    def __str__(self):
        return 'regexp %s' % self.regexp

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("regexp constraint doesn't apply to non "
                                      "final entity type")
        if not objschema in ('String', 'Password'):
            raise BadSchemaDefinition("regexp constraint doesn't apply to %s "
                                      "entity type" % objschema)

    def check(self, entity, rtype, value):
        """return true if the value maches the regular expression"""
        return self._rgx.match(value, self.flags)

    def failed_message(self, key, value):
        return _("%(KEY-value)r doesn't match the %(KEY-regexp)r regular expression"), {
            key+'-value': value,
            key+'-regexp': self.regexp}

    def serialize(self):
        """simple text serialization"""
        return u'%s,%s' % (self.regexp, self.flags)

    @classmethod
    def deserialize(cls, value):
        """simple text deserialization"""
        regexp, flags = value.rsplit(',', 1)
        return cls(regexp, int(flags))

    def __deepcopy__(self, memo):
        return RegexpConstraint(self.regexp, self.flags)


OPERATORS = {
    '<=': operator.le,
    '<': operator.lt,
    '>': operator.gt,
    '>=': operator.ge,
    }

class BoundaryConstraint(BaseConstraint):
    """the int/float bound constraint :

    set a minimal or maximal value to a numerical value
    """
    __implements__ = IConstraint

    def __init__(self, op, boundary=None):
        assert op in OPERATORS, op
        self.operator = op
        self.boundary = boundary

    def __str__(self):
        return 'value %s' % self.serialize()

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("bound constraint doesn't apply to non "
                                      "final entity type")

    def check(self, entity, rtype, value):
        """return true if the value satisfy the constraint, else false"""
        boundary = actual_value(self.boundary, entity)
        return OPERATORS[self.operator](value, boundary)

    def failed_message(self, key, value):
        return _("value %(KEY-value)s must be %(KEY-op)s %(KEY-boundary)s"), {
            key+'-value': value,
            key+'-op': self.operator,
            key+'-boundary': self.boundary}

    def serialize(self):
        """simple text serialization"""
        return u'%s %s' % (self.operator, self.boundary)

    @classmethod
    def deserialize(cls, value):
        """simple text deserialization"""
        op, boundary = value.split(' ', 1)
        return cls(op, eval(boundary))

BoundConstraint = class_renamed('BoundConstraint', BoundaryConstraint)
BoundConstraint.type = lambda x: 'BoundaryConstraint'


class IntervalBoundConstraint(BaseConstraint):
    """an int/float bound constraint :

    sets a minimal and / or a maximal value to a numerical value
    This class replaces the BoundConstraint class
    """
    __implements__ = IConstraint

    def __init__(self, minvalue=None, maxvalue=None):
        """
        :param minvalue: the minimal value that can be used
        :param maxvalue: the maxvalue value that can be used
        """
        assert not (minvalue is None and maxvalue is None)
        self.minvalue = minvalue
        self.maxvalue = maxvalue

    def __str__(self):
        return 'value [%s]' % self.serialize()

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("interval bound constraint doesn't apply "
                                      "to non final entity type")

    def check(self, entity, rtype, value):
        minvalue = actual_value(self.minvalue, entity)
        if minvalue is not None and value < minvalue:
            return False
        maxvalue = actual_value(self.maxvalue, entity)
        if maxvalue is not None and value > maxvalue:
            return False
        return True

    def failed_message(self, key, value):
        if self.minvalue is not None and value < self.minvalue:
            return _("value %(KEY-value)s must be >= %(KEY-boundary)s"), {
                key+'-value': value,
                key+'-boundary': self.minvalue}
        if self.maxvalue is not None and value > self.maxvalue:
            return _("value %(KEY-value)s must be <= %(KEY-boundary)s"), {
                key+'-value': value,
                key+'-boundary': self.maxvalue}
        assert False, 'shouldnt be there'

    def serialize(self):
        """simple text serialization"""
        return u'%s;%s' % (self.minvalue, self.maxvalue)

    @classmethod
    def deserialize(cls, value):
        """simple text deserialization"""
        minvalue, maxvalue = value.split(';')
        return cls(eval(minvalue), eval(maxvalue))


class StaticVocabularyConstraint(BaseConstraint):
    """Enforces a predefined vocabulary set for the value."""
    __implements__ = IVocabularyConstraint

    def __init__(self, values):
        self.values = tuple(values)

    def __str__(self):
        return 'value in (%s)' % self.serialize()

    def check(self, entity, rtype, value):
        """return true if the value is in the specific vocabulary"""
        return value in self.vocabulary(entity=entity)

    def failed_message(self, key, value):
        if isinstance(value, basestring):
            value = '"%s"' % unicode(value)
            choices = ', '.join('"%s"' % val for val in self.values)
        else:
            choices = ', '.join(unicode(val) for val in self.values)
        return _('invalid value %(KEY-value)s, it must be one of %(KEY-choices)s'), {
            key+'-value': value,
            key+'-choices': choices}

    def vocabulary(self, **kwargs):
        """return a list of possible values for the attribute"""
        return self.values

    def serialize(self):
        """serialize possible values as a csv list of evaluable strings"""
        try:
            sample = iter(self.vocabulary()).next()
        except:
            sample = unicode()
        if not isinstance(sample, basestring):
            return u', '.join(repr(word) for word in self.vocabulary())
        return u', '.join(repr(unicode(word).replace(',', ',,'))
                          for word in self.vocabulary())

    @classmethod
    def deserialize(cls, value):
        """deserialize possible values from a csv list of evaluable strings"""
        values = [eval(w) for w in re.split('(?<!,), ', value)]
        if values and isinstance(values[0], basestring):
            values = [v.replace(',,', ',') for v in values]
        return cls(values)


class FormatConstraint(StaticVocabularyConstraint):

    regular_formats = (_('text/rest'),
                       _('text/html'),
                       _('text/plain'),
                       )
    def __init__(self):
        self.values = self.vocabulary()

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("unique constraint doesn't apply to non "
                                      "final entity type")
        if not objschema == 'String':
            raise BadSchemaDefinition("format constraint only apply to String")

    def serialize(self):
        """called to make persistent valuable data of a constraint"""
        return None

    @classmethod
    def deserialize(cls, value):
        """called to restore serialized data of a constraint. Should return
        a `cls` instance
        """
        return cls()

    def vocabulary(self, **kwargs):
        return self.regular_formats

    def __str__(self):
        return 'value in (%s)' % u', '.join(repr(unicode(word)) for word in self.vocabulary())

FORMAT_CONSTRAINT = FormatConstraint()


class MultipleStaticVocabularyConstraint(StaticVocabularyConstraint):
    """Enforce a list of values to be in a predefined set vocabulary."""
    # XXX never used
    def check(self, entity, rtype, values):
        """return true if the values satisfy the constraint, else false"""
        vocab = self.vocabulary(entity=entity)
        for value in values:
            if not value in vocab:
                return False
        return True

# special classes to be used w/ constraints accepting values as argument(s):
# IntervalBoundConstraint

import datetime # needed for timedelta evaluation

def actual_value(value, entity):
    if hasattr(value, 'value'):
        return value.value(entity)
    return value


class Attribute(object):
    def __init__(self, attr):
        self.attr = attr

    def __str__(self):
        return '%s(%r)' % (self.__class__.__name__, self.attr)

    def value(self, entity):
        return getattr(entity, self.attr)


class NOW(object):
    def __init__(self, offset=None):
        self.offset = offset

    def __str__(self):
        return '%s(%r)' % (self.__class__.__name__, self.offset)

    def value(self, entity):
        now = yams.KEYWORD_MAP['Datetime']['NOW']()
        if self.offset:
            now += self.offset
        return now


class TODAY(object):
    def __init__(self, offset=None, type='Date'):
        self.offset = offset
        self.type = type

    def __str__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self.offset, self.type)

    def value(self, entity):
        now = yams.KEYWORD_MAP[self.type]['TODAY']()
        if self.offset:
            now += self.offset
        return now


# base types checking functions ################################################

def check_string(eschema, value):
    """check value is an unicode string"""
    return isinstance(value, unicode)

def check_password(eschema, value):
    """check value is an encoded string"""
    return isinstance(value, (str, StringIO))

def check_int(eschema, value):
    """check value is an integer"""
    try:
        int(value)
    except ValueError:
        return False
    return True

def check_float(eschema, value):
    """check value is a float"""
    try:
        float(value)
    except ValueError:
        return False
    return True

def check_decimal(eschema, value):
    """check value is a Decimal"""
    try:
        decimal.Decimal(value)
    except (TypeError, decimal.InvalidOperation):
        return False
    return True

def check_boolean(eschema, value):
    """check value is a boolean"""
    return isinstance(value, int)

def check_file(eschema, value):
    """check value has a getvalue() method (e.g. StringIO or cStringIO)"""
    return hasattr(value, 'getvalue')

def yes(*args, **kwargs):
    """dunno how to check"""
    return True


BASE_CHECKERS = {
    'Date' :     yes,
    'Time' :     yes,
    'Datetime' : yes,
    'TZTime' :   yes,
    'TZDatetime':yes,
    'Interval' : yes,
    'String' :   check_string,
    'Int' :      check_int,
    'BigInt' :   check_int,
    'Float' :    check_float,
    'Decimal' :  check_decimal,
    'Boolean' :  check_boolean,
    'Password' : check_password,
    'Bytes' :    check_file,
    }

BASE_CONVERTERS = {
    'String' :  unicode,
    'Password':  str,
    'Int' :      int,
    'BigInt' :   int,
    'Float' :    float,
    'Boolean' :  bool,
    'Decimal' :  decimal.Decimal,
    }

def patch_sqlite_decimal():
    """patch Decimal checker and converter to bypass SQLITE Bug
    (SUM of Decimal return float in SQLITE)"""
    def convert_decimal(value):
        # XXX issue a warning
        if isinstance(value, float):
            value = str(value)
        return decimal.Decimal(value)
    def check_decimal(eschema, value):
        """check value is a Decimal"""
        try:
            if isinstance(value, float):
                return True
            decimal.Decimal(value)
        except (TypeError, decimal.InvalidOperation):
            return False
        return True

    global BASE_CONVERTERS
    BASE_CONVERTERS['Decimal'] = convert_decimal
    global BASE_CHECKERS
    BASE_CHECKERS["Decimal"] = check_decimal



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Write a schema as a dot file.

"""
__docformat__ = "restructuredtext en"

import sys, os
import os.path as osp
from itertools import cycle

from logilab.common.graph import DotBackend, GraphGenerator

CARD_MAP = {'?': '0..1',
            '1': '1',
            '*': '0..n',
            '+': '1..n'}

class SchemaDotPropsHandler(object):
    def __init__(self, visitor):
        self.visitor = visitor
        # FIXME: colors are arbitrary
        self.nextcolor = cycle( ('#aa0000', '#00aa00', '#0000aa',
                                 '#000000', '#888888') ).next

    def node_properties(self, eschema):
        """return default DOT drawing options for an entity schema"""
        label = ['{',eschema.type,'|']
        label.append(r'\l'.join(rel.type for rel in eschema.ordered_relations()
                                if rel.final and self.visitor.should_display_attr(eschema, rel)))
        label.append(r'\l}') # trailing \l ensure alignement of the last one
        return {'label' : ''.join(label), 'shape' : "record",
                'fontname' : "Courier", 'style' : "filled"}

    def edge_properties(self, rschema, subjnode, objnode):
        """return default DOT drawing options for a relation schema"""
        # rschema can be none if the subject is a specialization of the object
        # we get there because we want to draw a specialization arrow anyway
        if rschema is None:
            kwargs = {'label': 'Parent class',
                      'color' : 'grey',  'style' : 'filled',
                      'arrowhead': 'empty'}
        elif rschema.symmetric:
            # symmetric rels are handled differently, let yams decide what's best
            kwargs = {'label': rschema.type,
                      'color': '#887788', 'style': 'dashed',
                      'dir': 'both', 'arrowhead': 'normal', 'arrowtail': 'normal'}
        else:
            kwargs = {'label': rschema.type,
                      'color' : 'black',  'style' : 'filled'}
            rdef = rschema.rdef(subjnode, objnode)
            composite = rdef.composite
            if rdef.composite == 'subject':
                kwargs['arrowhead'] = 'none'
                kwargs['arrowtail'] = 'diamond'
            elif rdef.composite == 'object':
                kwargs['arrowhead'] = 'diamond'
                kwargs['arrowtail'] = 'none'
            else:
                kwargs['arrowhead'] = 'normal'
                kwargs['arrowtail'] = 'none'
            # UML like cardinalities notation, omitting 1..1
            if rdef.cardinality[1] != '1':
                kwargs['taillabel'] = CARD_MAP[rdef.cardinality[1]]
            if rdef.cardinality[0] != '1':
                kwargs['headlabel'] = CARD_MAP[rdef.cardinality[0]]
            kwargs['color'] = self.nextcolor()
        kwargs['fontcolor'] = kwargs['color']
        # dot label decoration is just awful (1 line underlining the label
        # + 1 line going to the closest edge spline point)
        kwargs['decorate'] = 'false'
        #kwargs['labelfloat'] = 'true'
        return kwargs


class SchemaVisitor(object):
    def __init__(self, skiptypes=()):
        self._done = set()
        self.skiptypes = skiptypes
        self._nodes = None
        self._edges = None

    def should_display_schema(self, erschema):
        return not (erschema.final or erschema in self.skiptypes)

    def should_display_attr(self, eschema, rschema):
        return not rschema in self.skiptypes

    def display_rel(self, rschema, setype, tetype):
        if (rschema, setype, tetype) in self._done:
            return False
        self._done.add((rschema, setype, tetype))
        if rschema.symmetric:
            self._done.add((rschema, tetype, setype))
        return True

    def nodes(self):
        return self._nodes

    def edges(self):
        return self._edges


class FullSchemaVisitor(SchemaVisitor):
    def __init__(self, schema, skiptypes=()):
        super(FullSchemaVisitor, self).__init__(skiptypes)
        self.schema = schema
        self._eindex = None
        entities = [eschema for eschema in schema.entities()
                    if self.should_display_schema(eschema)]
        self._eindex = dict([(e.type, e) for e in entities])

    def nodes(self):
        for eschema in sorted(self._eindex.values(), key=lambda x: x.type):
            yield eschema.type, eschema

    def edges(self):
        # Entities with inheritance relations.
        for eschema in sorted(self._eindex.values(), key=lambda x: x.type):
            if eschema.specializes():
                yield str(eschema), str(eschema.specializes()), None
        # Subject/object relations.
        for rschema in sorted(self.schema.relations(), key=lambda x: x.type):
            if not self.should_display_schema(rschema):
                continue
            for setype, tetype in sorted(rschema.rdefs, key=lambda (s, o): (s.type, o.type)):
                if not (setype in self._eindex and tetype in self._eindex):
                    continue
                if not self.display_rel(rschema, setype, tetype):
                    continue
                yield str(setype), str(tetype), rschema


class OneHopESchemaVisitor(SchemaVisitor):
    def __init__(self, eschema, skiptypes=()):
        super(OneHopESchemaVisitor, self).__init__(skiptypes)
        nodes = set()
        edges = set()
        nodes.add((eschema.type, eschema))
        for rschema in eschema.subject_relations():
            if not self.should_display_schema(rschema):
                continue
            for teschema in rschema.objects(eschema.type):
                nodes.add((teschema.type, teschema))
                if not self.display_rel(rschema, eschema.type, teschema.type):
                    continue
                edges.add((eschema.type, teschema.type, rschema))
        for rschema in eschema.object_relations():
            if not self.should_display_schema(rschema):
                continue
            for teschema in rschema.subjects(eschema.type):
                nodes.add((teschema.type, teschema))
                if not self.display_rel(rschema, teschema.type, eschema.type):
                    continue
                edges.add((teschema.type, eschema.type, rschema))
        # Inheritance relations.
        if eschema.specializes():
            nodes.add((eschema.specializes().type, eschema.specializes()))
            edges.add((eschema.type, eschema.specializes().type, None))
        if eschema.specialized_by():
            for pschema in eschema.specialized_by():
                nodes.add((pschema.type, pschema))
                edges.add((pschema.type, eschema.type, None))
        self._nodes = nodes
        self._edges = edges


class OneHopRSchemaVisitor(SchemaVisitor):
    def __init__(self, rschema, skiptypes=()):
        super(OneHopRSchemaVisitor, self).__init__(skiptypes)
        nodes = set()
        edges = set()
        done = set()
        for seschema in rschema.subjects():
            nodes.add((seschema.type, seschema))
            for oeschema in rschema.objects(seschema.type):
                nodes.add((oeschema.type, oeschema))
                if not self.display_rel(rschema, seschema.type, oeschema.type):
                    continue
                edges.add((seschema.type, oeschema.type, rschema))
        self._nodes = nodes
        self._edges = edges


def schema2dot(schema=None, outputfile=None, skiptypes=(),
               visitor=None, prophdlr=None, size=None):
    """write to the output stream a dot graph representing the given schema"""
    visitor = visitor or FullSchemaVisitor(schema, skiptypes)
    prophdlr = prophdlr or SchemaDotPropsHandler(visitor)
    if outputfile:
        schemaname = osp.splitext(osp.basename(outputfile))[0]
    else:
        schemaname = 'Schema'
    generator = GraphGenerator(DotBackend(schemaname, 'BT',
                                          ratio='compress', size=size,
                                          renderer='dot',
                                          additionnal_param={
                                              'overlap':'false',
                                              'splines':'true',
                                              #'polylines':'true',
                                              'sep':'0.2'
                                          }))
    return generator.generate(visitor, prophdlr, outputfile)


def run():
    """main routine when schema2dot is used as a script"""
    from yams.reader import SchemaLoader
    loader = SchemaLoader()
    try:
        schema_dir = sys.argv[1]
    except IndexError:
        print "USAGE: schema2dot SCHEMA_DIR [OUTPUT FILE]"
        sys.exit(1)
    if len(sys.argv) > 2:
        outputfile = sys.argv[2]
    else:
        outputfile = None
    schema = loader.load([schema_dir], 'Test')
    schema2dot(schema, outputfile)


if __name__ == '__main__':
    run()



# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Compare two yams schemas

Textual representation of schema are created and standard diff algorithm are
applied.
"""

import subprocess
import tempfile
import os

from yams.constraints import (SizeConstraint,
                              UniqueConstraint,
                              StaticVocabularyConstraint)
from yams.reader import SchemaLoader


def quoted(obj):
    if isinstance(obj, basestring):
        # NOT repr, because .....
        # strings here turn out as unicode strings (from the repo side)
        # but are (mostly) really str in the schema.py source
        # and we don't want a bogus diff about '...' vs u'...'
        if "'" in obj:
            return '"%s"' % obj
    return "'%s'" % obj

class attr_scope(object):
    indentation = 2
    spacer = ''
    nextline = ','

class class_scope(object):
    indentation = 1
    spacer = ' '
    nextline = ''

def identity(x):
    return x

def format_props(props, scope, ignore=()):
    out = []
    for prop, val in sorted(props.iteritems()):
        if prop in ignore:
            continue
        out.append('\t' * scope.indentation +
                   '%s%s=%s%s' % (prop, scope.spacer, scope.spacer, val))
    return ('%s\n' % scope.nextline).join(out)

def format_expression(expr):
    mainvars = expr.mainvars - set('SUXO')
    return '%s(%s%s)' % (expr.__class__.__name__,
                         quoted(expr.expression),
                         ','.join(mainvars) if mainvars else '')

def format_tuple(iterator):
    out = []
    for thing in iterator:
        out.append(thing)
    return '(%s)' % ','.join(out)

def format_perms(perms, scope, isdefault):
    out = []
    for p in ('read', 'add', 'update', 'delete'):
        rule = perms.get(p)
        if rule:
            out.append('%s%s: %s' %
                       ('\t' + '\t' * scope.indentation,
                        quoted(p),
                        format_tuple(format_expression(r)
                                     if not isinstance(r, basestring) else quoted(r)
                              for r in rule)))
    return ' {%s\n%s\n%s}' % (' # default perms' if isdefault else '',
                              ',\n'.join(out),
                              '\t' * scope.indentation)

def format_constraint(cstr):
    cclass = cstr.__class__.__name__
    if 'RQL' in cclass:
        out = ['%s(%s' % (cclass, quoted(cstr.expression))]
        mainvars = getattr(cstr, 'mainvars', None)
        if mainvars:
            out.append(', mainvars=%s' % quoted(','.join(mainvars)))
        out.append(')')
    elif 'Boundary' in cclass:
        return '%s(%s %s)' % (cclass, quoted(cstr.operator), str(cstr.boundary))
    else:
        print 'formatter: unhandled constraint type', cstr.__class__
        return str(cstr)
    return ''.join(out)

def nullhandler(relation, perms):
    return perms, False

def properties_from(relation, permissionshandler=nullhandler):
    """return a dictionary containing properties of an attribute
    or a relation (if specified with is_relation option)"""
    ret = {}
    for prop in relation.rproperties():
        if prop == 'infered':
            continue

        val = getattr(relation, prop)
        if prop == 'permissions':
            val, default = permissionshandler(relation, val)
            if val:
                ret['__permissions__'] = format_perms(val, attr_scope, default)
            continue
        elif prop == 'cardinality' and relation.final:
            # attributes: re-sugar cardinality to 'required' prop
            prop = 'required'
            val = val[0] == '1'
            if not val:
                continue # don't emit 'required = False' props
        elif prop in ('indexed', 'fulltextindexed', 'uid', 'internationalizable'):
            if not val:
                continue # don't emit 'prop = False' props
        elif prop == 'constraints':
            other = []
            for constraint in val:
                # first, resugar some classic constraints
                if isinstance(constraint, SizeConstraint):
                    ret['maxsize'] = constraint.max
                elif isinstance(constraint, UniqueConstraint):
                    ret['unique'] = True
                elif isinstance(constraint, StaticVocabularyConstraint):
                    if relation.object.type == 'String':
                        transform = unicode
                    else:
                        transform = identity
                    ret['vocabulary'] = [transform(x)
                                         for x in constraint.vocabulary()]
                else:
                    other.append(constraint)
            if other:
                ret['constraints'] = '[%s]' % ','.join(format_constraint(cstr) for cstr in other)
            continue

        if val is None:
            continue

        if prop == 'default':
            if not isinstance(val, bool):
                if not val:
                    continue # default = ''
        elif prop in ('description', 'cardinality', 'composite'):
            if val:
                val = quoted(val)
            else: # let's skip empty strings
                continue

        ret[prop] = val
    return ret

def schema2descr(schema, permissionshandler, ignore=()):
    """convert a yams schema into a text description"""
    output = []
    ignore = set(ignore)
    multirtypes = []
    for entity in sorted(schema.entities()):
        if entity.final:
            continue
        output.append('class %s(EntityType):\n' % entity.type )
        perms, isdefault = permissionshandler(entity, entity.permissions)
        output.append('\t__permissions__ = %s\n\n' %format_perms(perms, class_scope, isdefault))

        attributes = [(attr[0].type,
                       attr[1].type,
                       properties_from(entity.rdef(attr[0].type), permissionshandler))
                      for attr in sorted(entity.attribute_definitions())]
        for attr_name, attr_type, attr_props in attributes:
            if attr_name in ignore:
                continue
            output.append('\t%s = %s(\n%s)\n\n' % (attr_name,
                                              attr_type,
                                              format_props(attr_props, attr_scope, ignore)))
        subject_relations = [(rel[0].type, # rtype
                              [r.type for r in rel[1]], # etype targets
                              properties_from(entity.rdef(rel[0].type), permissionshandler)) # properties
                             for rel in sorted(entity.relation_definitions())
                             if rel[2] == 'subject']
        for rtype, targettypes, props in subject_relations:
            targettypes = [etype for etype in targettypes
                           if etype not in ignore]
            if not targettypes:
                continue
            if len(targettypes) == 1:
                ttypes_out = quoted(targettypes[0])
            else:
                multirtypes.append(rtype)
                continue
            output.append('\t%s = SubjectRelation(%s\n%s)\n\n' %
                       (rtype, ttypes_out, format_props(props, attr_scope, ignore)) )
        output.append('\n\n')

    # Handle multi-rdef relation types
    for rtype in sorted(multirtypes):
        for (subjectschema, objectschema), relation in sorted(schema[rtype].rdefs.iteritems()):
            subjetype = subjectschema.type
            objetype = objectschema.type
            output.append('class %s_%s_%s(RelationDefinition):\n' %
                          (rtype, subjetype.lower(), objetype.lower()))
            output.append('\tname = %s\n' % quoted(rtype))
            output.append('\tsubject = %s\n' % quoted(subjetype))
            output.append('\tobject = %s\n' % quoted(objetype))
            output.append(format_props(properties_from(relation, permissionshandler),
                                       class_scope, ignore))
            output.append('\n\n')
    return ''.join(output)

def schema2file(schema, output, permissionshandler, ignore=()):
    """Save schema description of schema find
    in directory schema_dir into output file"""
    with open(output, 'wb') as description_file:
        description_file.write(schema2descr(schema, permissionshandler, ignore))

def schema_diff(schema1, schema2, permissionshandler=nullhandler, diff_tool=None, ignore=()):
    """schema 1 and 2 are CubicwebSchema
    diff_tool is the name of tool use for file comparison
    """
    tmpdir = tempfile.mkdtemp()
    output1 = os.path.join(tmpdir, "schema1.txt")
    output2 = os.path.join(tmpdir, "schema2.txt")
    schema2file(schema1, output1, permissionshandler, ignore)
    schema2file(schema2, output2, permissionshandler, ignore)
    if diff_tool:
        cmd = "%s %s %s" %(diff_tool,
                           output1, output2)
        process = subprocess.Popen(cmd, shell=True)
    else:
        print "description files save in %s and %s" % (output1, output2)
    print output1, output2
    return output1, output2

if __name__ == "__main__":
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option("-f", "--first-schema", dest="schema1",
                      help= "Specify the directory of the first schema")
    parser.add_option("-s", "--second-schema", dest="schema2",
                      help= "Specify the directory of the second schema")
    parser.add_option("-d", "--diff-tool", dest="diff_tool",
                      help= "Specify the name of the diff tool")
    (options, args) = parser.parse_args()
    if options.schema1 and options.schema2:
        schema1 = SchemaLoader().load([options.schema1])
        schema2 = SchemaLoader().load([options.schema2])
        output1, output2 = schema_diff(schema1, schema2)
        if options.diff_tool:
            cmd = "%s %s %s" %(options.diff_tool,
                               output1, output2)
            process = subprocess.Popen(cmd, shell=True)
        else:
            print "description files save in %s and %s" % (output1, output2)
    else:
        parser.error("An input file name must be specified")



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""specialization tests
"""
from logilab.common.testlib import TestCase, unittest_main

from yams.reader import build_schema_from_namespace
from yams.buildobjs import EntityType, String, SubjectRelation, RelationDefinition

def build_schema():

    class Person(EntityType):
        firstname = String()
        knows = SubjectRelation('Person')
        works_for = SubjectRelation('Company')

    class Student(Person):
        __specializes_schema__ = True

    class Company(EntityType):
        name = String()

    class SubCompany(Company):
        __specializes_schema__ = True

    class Division(Company):
        __specializes_schema__ = True
        division_of = SubjectRelation('Company')

    class SubDivision(Division):
        __specializes_schema__ = True

    # This class doesn't extend the schema
    class SubSubDivision(SubDivision):
        pass

    class custom_attr(RelationDefinition):
        subject = 'Person'
        object = 'String'
        __permissions__ = {'read': ('managers', ),
                           'add': ('managers', ),
                           'update': ('managers', )}

    return build_schema_from_namespace(locals().items())


class SpecializationTC(TestCase):
    def setUp(self):
        self.schema = build_schema()

    def test_schema_specialization(self):
        schema = self.schema
        # company
        company = schema.eschema('Company')
        self.assertEqual(company.specializes(), None)
        # division
        division = schema.eschema('Division')
        self.assertEqual(division.specializes().type, 'Company')
        # subdivision
        subdivision = schema.eschema('SubDivision')
        self.assertEqual(subdivision.specializes().type, 'Division')
        # subsubdivision
        subsubdivision = schema.eschema('SubSubDivision')
        self.assertEqual(subsubdivision.specializes(), None)

    def test_ancestors(self):
        schema = self.schema
        # company
        company = schema.eschema('Company')
        self.assertEqual(company.ancestors(), [])
        # division
        division = schema.eschema('Division')
        self.assertEqual(division.ancestors(), ['Company'])
        # subdivision
        subdivision = schema.eschema('SubDivision')
        self.assertEqual(subdivision.ancestors(), ['Division', 'Company'])
        # subsubdivision
        subsubdivision = schema.eschema('SubSubDivision')
        self.assertEqual(subsubdivision.ancestors(), [])

    def test_specialized_by(self):
        schema = self.schema
        # company
        company = schema.eschema('Company')
        self.assertEqual(sorted(company.specialized_by(False)), ['Division', 'SubCompany'])
        self.assertEqual(sorted(company.specialized_by(True)), ['Division', 'SubCompany', 'SubDivision'])
        # division
        division = schema.eschema('Division')
        self.assertEqual(sorted(division.specialized_by(False)), ['SubDivision'])
        self.assertEqual(sorted(division.specialized_by(True)), ['SubDivision'])
        # subdivision
        subdivision = schema.eschema('SubDivision')
        self.assertEqual(sorted(subdivision.specialized_by(False)), [])
        # subsubdivision
        subsubdivision = schema.eschema('SubSubDivision')
        self.assertEqual(subsubdivision.specialized_by(False), [])

    def test_relations_infered(self):
        entities = [str(e) for e in self.schema.entities() if not e.final]
        relations = sorted([r for r in self.schema.relations() if not r.final])
        self.assertListEqual(sorted(entities), ['Company', 'Division', 'Person',
                                                 'Student', 'SubCompany', 'SubDivision', 'SubSubDivision'])
        self.assertListEqual(relations, ['division_of', 'knows', 'works_for'])
        expected = {('Person', 'Person'): False,
                    ('Person', 'Student'): True,
                    # as Student extends Person, it already has the `knows` relation
                    ('Student', 'Person'): False,
                    ('Student', 'Student'): True,
                    }
        done = set()
        drschema, krschema, wrschema = relations
        for subjobj in krschema.rdefs:
            subject, object = subjobj
            done.add(subjobj)
            self.assertTrue(subjobj in expected)
            self.assertEqual(krschema.rdef(subject, object).infered,
                              expected[subjobj])
        self.assertEqual(len(set(expected) - done), 0, 'missing %s' % (set(expected) - done))
        expected = {('Person', 'Company'): False,
                    ('Person', 'Division'): True,
                    ('Person', 'SubDivision'): True,
                    ('Person', 'SubCompany'): True,
                    ('Student', 'Company'): False,
                    ('Student', 'Division'): True,
                    ('Student', 'SubDivision'): True,
                    ('Student', 'SubCompany'): True,
                    }
        done = set()
        for subjobj in wrschema.rdefs:
            subject, object = subjobj
            done.add(subjobj)
            self.assertTrue(subjobj in expected)
            self.assertEqual(wrschema.rdef(subject, object).infered,
                              expected[subjobj])
        self.assertEqual(len(set(expected) - done), 0, 'missing %s' % (set(expected) - done))

        self.assertIn('custom_attr', self.schema['Student'].subjrels)
        self.assertEqual(
            self.schema['custom_attr'].rdefs[('Student', 'String')].permissions,
            {'read': ('managers',),
             'add': ('managers',),
             'update': ('managers',)})

    def test_remove_infered_relations(self):
        self.schema.remove_infered_definitions()
        relations = sorted([r for r in self.schema.relations() if not r.final])
        self.assertListEqual(relations, ['division_of', 'knows', 'works_for'])
        expected = {('Person', 'Person'): False,
                    # as Student extends Person, it already has the `knows` relation
                    ('Student', 'Person'): False,
                    }
        done = set()
        drschema, krschema, wrschema = relations
        for subjobj in krschema.rdefs:
            subject, object = subjobj
            done.add(subjobj)
            self.assertTrue(subjobj in expected)
            self.assertEqual(krschema.rdef(subject, object).infered,
                              expected[subjobj])
        self.assertEqual(len(set(expected) - done), 0, 'missing %s' % (set(expected) - done))
        expected = {('Person', 'Company'): False,
                    ('Student', 'Company'): False,
                   }
        done = set()
        for subjobj in wrschema.rdefs:
            subject, object = subjobj
            done.add(subjobj)
            self.assertTrue(subjobj in expected)
            self.assertEqual(wrschema.rdef(subject, object).infered,
                              expected[subjobj])
        self.assertEqual(len(set(expected) - done), 0, 'missing %s' % (set(expected) - done))

    def test_no_more_infered_relations(self):
        rdef = self.schema['division_of'].rdefs['SubSubDivision', 'SubCompany']
        self.assertEqual('**', rdef.cardinality)
        rdef = RelationDefinition('SubSubDivision', 'division_of', 'SubCompany',
                                  cardinality='1*')
        # ensure add_relation_def doesn't raise an error
        self.schema.add_relation_def(rdef)
        rdef = self.schema['division_of'].rdefs['SubSubDivision', 'SubCompany']
        self.assertEqual('1*', rdef.cardinality)
        rdef = self.schema['SubSubDivision'].rdef('division_of', 'subject', 'SubCompany')
        self.assertEqual('1*', rdef.cardinality)
        rdef = self.schema['SubCompany'].rdef('division_of', 'object', 'SubSubDivision')
        self.assertEqual('1*', rdef.cardinality)

if __name__ == '__main__':
    unittest_main()



import logging
from cStringIO import StringIO
from logilab.common.graph import ordered_nodes

def serialize_to_python(s):
    out = StringIO()
    w = out.write
    w('from yams.buildobjs import *\n\n')
    graph = {}
    for entity in s.entities():
        l = graph.setdefault(entity, [])
        if entity._specialized_type is not None:
            l.append(entity._specialized_type)
    for e in reversed(ordered_nodes(graph)):
        if not e.final:
            if e._specialized_type:
                base = e._specialized_type
            else:
                base = 'EntityType'
            w('class %s(%s):\n' % (e.type, base))
            attr_defs = list(e.attribute_definitions())
            if attr_defs:
                for attr,obj in attr_defs:
                    w('    %s = %s()\n' % (attr.type, obj.type))
            else:
                w('    pass\n')
            w('\n')
    for r in s.relations():
        if not r.final:
            if r.subjects() and r.objects():
                w('class %s(RelationDefinition):\n' % r.type)
                w('    subject = (%s,)\n' % ', '.join("'%s'" % x for x in r.subjects()))
                w('    object = (%s,)\n' % ', '.join("'%s'" % x for x in r.objects()))
                w('\n')
            else:
                logging.warning('relation definition %s missing subject/object' % r.type)
    return out.getvalue()



# copyright 2004-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""write a schema as sql"""

__docformat__ = "restructuredtext en"

from hashlib import md5

from yams.constraints import SizeConstraint, UniqueConstraint

# default are usually not handled at the sql level. If you want them, set
# SET_DEFAULT to True
SET_DEFAULT = False


def schema2sql(dbhelper, schema, skip_entities=(), skip_relations=(), prefix=''):
    """write to the output stream a SQL schema to store the objects
    corresponding to the given schema
    """
    output = []
    w = output.append
    for etype in sorted(schema.entities()):
        eschema = schema.eschema(etype)
        if eschema.final or eschema.type in skip_entities:
            continue
        w(eschema2sql(dbhelper, eschema, skip_relations, prefix=prefix))
    for rtype in sorted(schema.relations()):
        rschema = schema.rschema(rtype)
        if rschema.final or rschema.inlined:
            continue
        w(rschema2sql(rschema))
    return '\n'.join(output)


def dropschema2sql(dbhelper, schema, skip_entities=(), skip_relations=(), prefix=''):
    """write to the output stream a SQL schema to store the objects
    corresponding to the given schema
    """
    output = []
    w = output.append
    for etype in sorted(schema.entities()):
        eschema = schema.eschema(etype)
        if eschema.final or eschema.type in skip_entities:
            continue
        stmts = dropeschema2sql(dbhelper, eschema, skip_relations, prefix=prefix)
        for stmt in stmts:
            w(stmt)
    for rtype in sorted(schema.relations()):
        rschema = schema.rschema(rtype)
        if rschema.final or rschema.inlined:
            continue
        w(droprschema2sql(rschema))
    return '\n'.join(output)


def eschema_attrs(eschema, skip_relations):
    attrs = [attrdef for attrdef in eschema.attribute_definitions()
             if not attrdef[0].type in skip_relations]
    attrs += [(rschema, None)
              for rschema in eschema.subject_relations()
              if not rschema.final and rschema.inlined]
    return attrs

def unique_index_name(eschema, columns):
    return u'unique_%s' % md5(eschema.type +
                              ',' +
                              ','.join(sorted(columns))).hexdigest()

def iter_unique_index_names(eschema):
    for columns in eschema._unique_together or ():
        yield columns, unique_index_name(eschema, columns)

def dropeschema2sql(dbhelper, eschema, skip_relations=(), prefix=''):
    """return sql to drop an entity type's table"""
    # not necessary to drop indexes, that's implictly done when
    # dropping the table, but we need to drop SQLServer views used to
    # create multicol unique indices
    statements = []
    tablename = prefix + eschema.type
    if eschema._unique_together is not None:
        for columns, index_name in iter_unique_index_names(eschema):
            cols  = ['%s%s' % (prefix, col) for col in columns]
            sqls = dbhelper.sqls_drop_multicol_unique_index(tablename, cols, index_name)
            statements += sqls
    statements += ['DROP TABLE %s;' % (tablename)]
    return statements


def eschema2sql(dbhelper, eschema, skip_relations=(), prefix=''):
    """write an entity schema as SQL statements to stdout"""
    output = []
    w = output.append
    table = prefix + eschema.type
    w('CREATE TABLE %s(' % (table))
    attrs = eschema_attrs(eschema, skip_relations)
    # XXX handle objectinline physical mode
    for i in xrange(len(attrs)):
        rschema, attrschema = attrs[i]
        if attrschema is not None:
            sqltype = aschema2sql(dbhelper, eschema, rschema, attrschema,
                                  indent=' ')
        else: # inline relation
            # XXX integer is ginco specific
            sqltype = 'integer'
        if i == len(attrs) - 1:
            w(' %s%s %s' % (prefix, rschema.type, sqltype))
        else:
            w(' %s%s %s,' % (prefix, rschema.type, sqltype))
    w(');')
    # create indexes
    for i in xrange(len(attrs)):
        rschema, attrschema = attrs[i]
        if attrschema is None or eschema.rdef(rschema).indexed:
            w(dbhelper.sql_create_index(table, prefix + rschema.type))
    for columns, index_name in iter_unique_index_names(eschema):
        cols  = ['%s%s' % (prefix, col) for col in columns]
        sqls = dbhelper.sqls_create_multicol_unique_index(table, cols, index_name)
        for sql in sqls:
            w(sql)
    w('')
    return '\n'.join(output)


def aschema2sql(dbhelper, eschema, rschema, aschema, creating=True, indent=''):
    """write an attribute schema as SQL statements to stdout"""
    attr = rschema.type
    rdef = rschema.rdef(eschema.type, aschema.type)
    sqltype = type_from_constraints(dbhelper, aschema.type, rdef.constraints,
                                    creating)
    if SET_DEFAULT:
        default = eschema.default(attr)
        if default is not None:
            if aschema.type == 'Boolean':
                sqltype += ' DEFAULT %s' % dbhelper.boolean_value(default)
            elif aschema.type == 'String':
                sqltype += ' DEFAULT %r' % str(default)
            elif aschema.type in ('Int', 'BigInt', 'Float'):
                sqltype += ' DEFAULT %s' % default
            # XXX ignore default for other type
            # this is expected for NOW / TODAY
    if creating:
        if rdef.uid:
            sqltype += ' PRIMARY KEY'
        elif rdef.cardinality[0] == '1':
            # don't set NOT NULL if backend isn't able to change it later
            if dbhelper.alter_column_support:
                sqltype += ' NOT NULL'
    # else we're getting sql type to alter a column, we don't want key / indexes
    # / null modifiers
    return sqltype


def type_from_constraints(dbhelper, etype, constraints, creating=True):
    """return a sql type string corresponding to the constraints"""
    constraints = list(constraints)
    unique, sqltype = False, None
    size_constrained_string = dbhelper.TYPE_MAPPING.get('SizeConstrainedString', 'varchar(%s)')
    if etype == 'String':
        for constraint in constraints:
            if isinstance(constraint, SizeConstraint):
                if constraint.max is not None:
                    sqltype = size_constrained_string % constraint.max
            elif isinstance(constraint, UniqueConstraint):
                unique = True
    if sqltype is None:
        sqltype = dbhelper.TYPE_MAPPING[etype]
    if creating and unique:
        sqltype += ' UNIQUE'
    return sqltype


_SQL_SCHEMA = """
CREATE TABLE %(table)s (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT %(table)s_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX %(table)s_from_idx ON %(table)s(eid_from);
CREATE INDEX %(table)s_to_idx ON %(table)s(eid_to);"""


def rschema2sql(rschema):
    return _SQL_SCHEMA % {'table': '%s_relation' % rschema.type}


def droprschema2sql(rschema):
    """return sql to drop a relation type's table"""
    # not necessary to drop indexes, that's implictly done when dropping
    # the table
    return 'DROP TABLE %s_relation;' % rschema.type


def grant_schema(schema, user, set_owner=True, skip_entities=(), prefix=''):
    """write to the output stream a SQL schema to store the objects
    corresponding to the given schema
    """
    output = []
    w = output.append
    for etype in sorted(schema.entities()):
        eschema = schema.eschema(etype)
        if eschema.final or etype in skip_entities:
            continue
        w(grant_eschema(eschema, user, set_owner, prefix=prefix))
    for rtype in sorted(schema.relations()):
        rschema = schema.rschema(rtype)
        if rschema.final or rschema.inlined:
            continue
        w(grant_rschema(rschema, user, set_owner))
    return '\n'.join(output)


def grant_eschema(eschema, user, set_owner=True, prefix=''):
    output = []
    w = output.append
    etype = eschema.type
    if set_owner:
        w('ALTER TABLE %s%s OWNER TO %s;' % (prefix, etype, user))
    w('GRANT ALL ON %s%s TO %s;' % (prefix, etype, user))
    return '\n'.join(output)


def grant_rschema(rschema, user, set_owner=True):
    output = []
    if set_owner:
        output.append('ALTER TABLE %s_relation OWNER TO %s;' % (rschema.type, user))
    output.append('GRANT ALL ON %s_relation TO %s;' % (rschema.type, user))
    return '\n'.join(output)



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""unit tests for module yams.reader"""

from __future__ import with_statement

import sys
import os.path as osp
from datetime import datetime, date, time

from logilab.common.testlib import TestCase, unittest_main

from yams import BadSchemaDefinition
from yams.schema import Schema
from yams.reader import SchemaLoader, build_schema_from_namespace
from yams.constraints import StaticVocabularyConstraint, SizeConstraint
from yams.buildobjs import (EntityType, RelationType, RelationDefinition,
                            SubjectRelation,
                            Int, String, Float, Datetime, Date, Boolean,
                            DEFAULT_RELPERMS, DEFAULT_ATTRPERMS)

sys.path.insert(0, osp.dirname(__file__))


class SchemaLoaderTC(TestCase):

    # test helper functions ###################################################

    def test_get_schema_files(self):
        files = [osp.basename(f) for f in SchemaLoader().get_schema_files(self.datadir)]
        self.assertEqual(files[0], '__init__.py')
        self.assertEqual(sorted(files),
                          ['Company.py', 'Dates.py', 'State.py', '__init__.py', 'schema.py'])

    # test load_schema read entity and relation types #######################

    def test_load_schema(self):
        self.assert_(isinstance(SCHEMA, Schema))
        self.assertEqual(SCHEMA.name, 'NoName')
        self.assertListEqual(sorted(SCHEMA.entities()),
                              ['Affaire', 'BigInt', 'Boolean', 'Bytes', 'Company',
                               'Date', 'Datetest', 'Datetime', 'Decimal',
                               'Division', 'EPermission', 'Eetype',  'Employee', 'Float', 'Int', 'Interval',
                               'Note', 'Password', 'Person', 'Salaried', 'Societe', 'State', 'String',
                               'Subcompany', 'Subdivision', 'TZDatetime', 'TZTime', 'Time', 'pkginfo'])
        self.assertListEqual(sorted(SCHEMA.relations()),
                              ['ad1', 'ad2', 'ad3', 'adel', 'ass', 'author', 'author_email',
                               'concerne', 'copyright', 'cp',
                               'd1', 'd2', 'date', 'datenaiss', 'debian_handler', 'description', 'division_of', 'dt1', 'dt2',
                               'eid', 'evaluee', 'fax', 'final',
                               'initial_state', 'inline_rel',
                               'license', 'long_desc',
                               'mailinglist', 'meta', 'modname',
                               'name', 'next_state', 'nom', 'obj_wildcard',
                               'para', 'prenom', 'promo',
                               'ref', 'require_permission', 'rncs',
                               'salary', 'sexe', 'short_desc', 'state_of', 'subcompany_of',
                               'subdivision_of', 'subj_wildcard', 'sujet', 'sym_rel',
                               't1', 't2', 'tel', 'test', 'titre', 'travaille', 'type',
                               'version',
                               'ville', 'web', 'works_for'])

    def test_eschema(self):
        eschema = SCHEMA.eschema('Societe')
        self.assertEqual(eschema.description, '')
        self.assertEqual(eschema.final, False)
        self.assertListEqual(sorted(eschema.subject_relations()),
                              ['ad1', 'ad2', 'ad3', 'cp', 'evaluee',
                               'fax', 'nom', 'rncs', 'subj_wildcard', 'tel', 'ville',
                               'web'])
        self.assertListEqual(sorted(eschema.object_relations()),
                          ['concerne', 'obj_wildcard', 'travaille'])

        eschema = SCHEMA.eschema('Eetype')
        self.assertEqual(eschema.description, 'define an entity type, used to build the application schema')
        self.assertEqual(eschema.final, False)
        self.assertListEqual(sorted(str(r) for r in eschema.subject_relations()),
                              ['description', 'final', 'initial_state', 'meta',
                               'name', 'subj_wildcard'])
        self.assertListEqual(sorted(str(r) for r in eschema.object_relations()),
                              ['obj_wildcard', 'state_of'])

        eschema = SCHEMA.eschema('Boolean')
        self.assertEqual(eschema.description, '')
        self.assertEqual(eschema.final, True)
        self.assertListEqual(sorted(eschema.subject_relations()),
                              [])
        self.assertListEqual(sorted(eschema.object_relations()),
                              ['final', 'meta', 'test'])

    # test base entity type's subject relation properties #####################

    def test_indexed(self):
        eschema = SCHEMA.eschema('Person')
        self.assert_(not eschema.rdef('nom').indexed)
        eschema = SCHEMA.eschema('State')
        self.assert_(eschema.rdef('name').indexed)

    def test_uid(self):
        eschema = SCHEMA.eschema('State')
        self.assert_(eschema.rdef('eid').uid)
        self.assert_(not eschema.rdef('name').uid)

    def test_fulltextindexed(self):
        eschema = SCHEMA.eschema('Person')
        self.assertRaises(AttributeError, getattr, eschema.rdef('tel'), 'fulltextindexed') # tel is an INT
        self.assert_(eschema.rdef('nom').fulltextindexed)
        self.assert_(eschema.rdef('prenom').fulltextindexed)
        self.assert_(not eschema.rdef('sexe').fulltextindexed)
        indexable = sorted(eschema.indexable_attributes())
        self.assertEqual(['nom', 'prenom', 'titre'], indexable)
        self.assertEqual(SCHEMA.rschema('works_for').fulltext_container, None)
        self.assertEqual(SCHEMA.rschema('require_permission').fulltext_container,
                          'subject')
        eschema = SCHEMA.eschema('Company')
        indexable = sorted(eschema.indexable_attributes())
        self.assertEqual([], indexable)
        indexable = sorted(eschema.fulltext_relations())
        self.assertEqual([('require_permission', 'subject')], indexable)
        containers = sorted(eschema.fulltext_containers())
        self.assertEqual([], containers)
        eschema = SCHEMA.eschema('EPermission')
        indexable = sorted(eschema.indexable_attributes())
        self.assertEqual(['name'], indexable)
        indexable = sorted(eschema.fulltext_relations())
        self.assertEqual([], indexable)
        containers = sorted(eschema.fulltext_containers())
        self.assertEqual([('require_permission', 'subject')], containers)

    def test_internationalizable(self):
        eschema = SCHEMA.eschema('Eetype')
        self.assert_(eschema.rdef('name').internationalizable)
        eschema = SCHEMA.eschema('State')
        self.assert_(eschema.rdef('name').internationalizable)
        eschema = SCHEMA.eschema('Societe')
        self.assert_(not eschema.rdef('ad1').internationalizable)

    # test advanced entity type's subject relation properties #################

    def test_vocabulary(self):
        eschema = SCHEMA.eschema('pkginfo')
        self.assertEqual(eschema.vocabulary('license'), ('GPL', 'ZPL'))
        self.assertEqual(eschema.vocabulary('debian_handler'), ('machin', 'bidule'))

    def test_default(self):
        eschema = SCHEMA.eschema('pkginfo')
        self.assertEqual(eschema.default('version'), '0.1')
        self.assertEqual(eschema.default('license'), None)

    # test relation type properties ###########################################

    def test_rschema(self):
        rschema = SCHEMA.rschema('evaluee')
        self.assertEqual(rschema.symmetric, False)
        self.assertEqual(rschema.description, '')
        self.assertEqual(rschema.final, False)
        self.assertListEqual(sorted(rschema.subjects()), ['Person', 'Salaried', 'Societe'])
        self.assertListEqual(sorted(rschema.objects()), ['Note'])

        rschema = SCHEMA.rschema('sym_rel')
        self.assertEqual(rschema.symmetric, True)
        self.assertEqual(rschema.description, '')
        self.assertEqual(rschema.final, False)
        self.assertListEqual(sorted(rschema.subjects()), ['Affaire', 'Person', 'Salaried'])
        self.assertListEqual(sorted(rschema.objects()), ['Affaire', 'Person', 'Salaried'])

        rschema = SCHEMA.rschema('initial_state')
        self.assertEqual(rschema.symmetric, False)
        self.assertEqual(rschema.description, 'indicate which state should be used by default when an entity using states is created')
        self.assertEqual(rschema.final, False)
        self.assertListEqual(sorted(rschema.subjects()), ['Eetype'])
        self.assertListEqual(sorted(rschema.objects()), ['State'])

        rschema = SCHEMA.rschema('name')
        self.assertEqual(rschema.symmetric, False)
        self.assertEqual(rschema.description, '')
        self.assertEqual(rschema.final, True)
        self.assertListEqual(sorted(rschema.subjects()), ['Company', 'Division', 'EPermission', 'Eetype', 'State', 'Subcompany', 'Subdivision'])
        self.assertListEqual(sorted(rschema.objects()), ['String'])

    def test_cardinality(self):
        rschema = SCHEMA.rschema('evaluee')
        self.assertEqual(rschema.rdef('Person', 'Note').cardinality, '**')
        rschema = SCHEMA.rschema('inline_rel')
        self.assertEqual(rschema.rdef('Affaire', 'Person').cardinality, '?*')
        rschema = SCHEMA.rschema('initial_state')
        self.assertEqual(rschema.rdef('Eetype', 'State').cardinality, '?*')
        rschema = SCHEMA.rschema('state_of')
        self.assertEqual(rschema.rdef('State', 'Eetype').cardinality, '+*')
        rschema = SCHEMA.rschema('name')
        self.assertEqual(rschema.rdef('State', 'String').cardinality, '11')
        rschema = SCHEMA.rschema('description')
        self.assertEqual(rschema.rdef('State', 'String').cardinality, '?1')

    def test_constraints(self):
        eschema = SCHEMA.eschema('Person')
        self.assertEqual(len(eschema.rdef('nom').constraints), 1)
        self.assertEqual(len(eschema.rdef('promo').constraints), 2)
        self.assertEqual(len(eschema.rdef('tel').constraints), 0)
        eschema = SCHEMA.eschema('State')
        self.assertEqual(len(eschema.rdef('name').constraints), 1)
        self.assertEqual(len(eschema.rdef('description').constraints), 0)
        eschema = SCHEMA.eschema('Eetype')
        self.assertEqual(len(eschema.rdef('name').constraints), 2)

    def test_inlined(self):
        rschema = SCHEMA.rschema('evaluee')
        self.assertEqual(rschema.inlined, False)
        rschema = SCHEMA.rschema('state_of')
        self.assertEqual(rschema.inlined, False)
        rschema = SCHEMA.rschema('inline_rel')
        self.assertEqual(rschema.inlined, True)
        rschema = SCHEMA.rschema('initial_state')
        self.assertEqual(rschema.inlined, True)

    def test_relation_permissions(self):
        rschema = SCHEMA.rschema('evaluee')
        self.assertEqual(rschema.rdef('Person', 'Note').permissions,
                          {'read': ('managers',),
                           'delete': ('managers',),
                           'add': ('managers',)})
        self.assertEqual(rschema.rdef('Societe', 'Note').permissions,
                          {'read': ('managers',),
                           'delete': ('managers',),
                           'add': ('managers',)})
        rschema = SCHEMA.rschema('concerne')
        self.assertEqual(rschema.rdef('Person', 'Affaire').permissions,
                          {'read': ('managers',),
                           'delete': ('managers',),
                           'add': ('managers',)})
        self.assertEqual(rschema.rdef('Affaire', 'Societe').permissions,
                          DEFAULT_RELPERMS)
        rschema = SCHEMA.rschema('travaille')
        self.assertEqual(rschema.rdef('Person', 'Societe').permissions,
                          {'read': (), 'add': (), 'delete': ('managers',)})

    def test_attributes_permissions(self):
        rschema = SCHEMA.rschema('name')
        self.assertEqual(rschema.rdef('Company', 'String').permissions,
                          DEFAULT_ATTRPERMS)
        rschema = SCHEMA.rschema('tel')
        self.assertEqual(rschema.rdef('Person', 'Int').permissions,
                          {'read': (),
                           'add': ('managers',),
                           'update': ('managers',)})


    def test_entity_permissions(self):
        eschema = SCHEMA.eschema('State')
        self.assertEqual(eschema.permissions,
                          {'read':   ('managers', 'users', 'guests',),
                           'add':    ('managers', 'users',),
                           'delete': ('managers', 'owners',),
                           'update': ('managers', 'owners',)})

        eschema = SCHEMA.eschema('Eetype')
        self.assertEqual(eschema.permissions,
                          {'read':   ('managers', 'users', 'guests',),
                           'add':    ('managers',),
                           'delete': ('managers',),
                           'update': ('managers', 'owners',)})

        eschema = SCHEMA.eschema('Person')
        self.assertEqual(eschema.permissions,
                          {'read':   ('managers', 'users', 'guests',),
                           'add':    ('managers', 'users',),
                           'delete': ('managers', 'owners',),
                           'update': ('managers', 'owners',)})

##     def test_nonregr_using_tuple_as_relation_target(self):
##         rschema = schema.rschema('see_also')
##         self.assertEqual(rschema.symmetric, False)
##         self.assertEqual(rschema.description, '')
##         self.assertEqual(rschema.final, False)
##         self.assertListEqual(sorted(rschema.subjects()), ['Employee'])
##         self.assertListEqual(sorted(rschema.objects()), ['Company', 'Division'])
##

SCHEMA = SchemaLoader().load([SchemaLoaderTC.datadir])


class BasePerson(EntityType):
    firstname = String(vocabulary=('logilab', 'caesium'), maxsize=10)
    lastname = String(constraints=[StaticVocabularyConstraint(['logilab', 'caesium'])])

class Person(BasePerson):
    email = String()

class Employee(Person):
    company = String(vocabulary=('logilab', 'caesium'))


class Student(Person):
    __specializes_schema__ = True
    college = String()

class X(Student):
    pass

class Foo(EntityType):
    i = Int(required=True, metadata={'name': String()})
    f = Float()
    d = Datetime()



class PySchemaTC(TestCase):

    def test_python_inheritance(self):
        bp = BasePerson()
        p = Person()
        e = Employee()
        self.assertEqual([r.name for r in bp.__relations__], ['firstname', 'lastname'])
        self.assertEqual([r.name for r in p.__relations__], ['firstname', 'lastname', 'email'])
        self.assertEqual([r.name for r in e.__relations__], ['firstname', 'lastname', 'email', 'company'])

    def test_schema_extension(self):
        s = Student()
        self.assertEqual([r.name for r in s.__relations__], ['firstname', 'lastname', 'email', 'college'])
        self.assertEqual(s.specialized_type, 'Person')
        x = X()
        self.assertEqual(x.specialized_type, None)

    def test_relationtype(self):
        foo = Foo()
        self.assertEqual(['Int', 'String', 'Float', 'Datetime'],
                         [r.etype for r in foo.__relations__])
        self.assertEqual(foo.__relations__[0].cardinality, '11')
        self.assertEqual(foo.__relations__[2].cardinality, '?1')

    def test_maxsize(self):
        bp = BasePerson()
        def maxsize(e):
            for e in e.constraints:
                if isinstance(e, SizeConstraint):
                    return e.max
        self.assertEqual(maxsize(bp.__relations__[0]), 7)
        # self.assertEqual(maxsize(bp.__relations__[1]), 7)
        emp = Employee()
        self.assertEqual(maxsize(emp.__relations__[3]), 7)

    def test_metadata(self):
        foo = Foo()
        self.assertEqual('i_name', foo.__relations__[1].name)


    def test_date_defaults(self):
        _today = date.today()
        _now = datetime.now()
        datetest = SCHEMA.eschema('Datetest')
        dt1 = datetest.default('dt1')
        dt2 = datetest.default('dt2')
        d1 = datetest.default('d1')
        d2 = datetest.default('d2')
        t1 = datetest.default('t1')
        t2 = datetest.default('t2')
        # datetimes
        self.assertIsInstance(dt1, datetime)
        # there's no easy way to test NOW (except monkey patching now() itself)
        delta = dt1 - _now
        self.assertTrue(abs(delta.seconds) < 5)
        self.assertEqual(date(dt2.year, dt2.month, dt2.day), _today)
        self.assertIsInstance(dt2, datetime)
        # dates
        self.assertEqual(d1, _today)
        self.assertIsInstance(d1, date)
        self.assertEqual(d2, date(2007, 12, 11))
        self.assertIsInstance(d2, date)
        # times
        self.assertEqual(t1, time(8, 40))
        self.assertIsInstance(t1, time)
        self.assertEqual(t2, time(9, 45))
        self.assertIsInstance(t2, time)


class SchemaLoaderTC2(TestCase):
    def tearDown(self):
        SchemaLoader.main_schema_directory = 'schema'

    def assertBadInlinedMessage(self, error):
        try:
            self.assertEqual("conflicting values False/True for property inlined of relation 'rel'",
                             str(error))
        except AssertionError:
            self.assertEqual("conflicting values True/False for property inlined of relation 'rel'",
                             str(error))

    def test_bad_rtype_inlined_conflict1(self):
        class Anentity(EntityType):
            rel = SubjectRelation('Anentity', inlined=True)
        class Anotherentity(EntityType):
            rel = SubjectRelation('Anentity', inlined=False)

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(locals().items())
        self.assertBadInlinedMessage(cm.exception)


    def test_bad_rtype_inlined_conflict2(self):
        class Anentity(EntityType):
            rel = SubjectRelation('Anentity', inlined=True)
        class rel(RelationType):
            inlined = False

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(locals().items())
        self.assertBadInlinedMessage(cm.exception)

    def test_bad_int_size_constraint(self):
        class Entity(EntityType):
            attr = Int(maxsize=40)

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(locals().items())
        self.assertEqual("size constraint doesn't apply to Int entity type",
                         str(cm.exception))

    def test_bad_vocab_and_size(self):
        class Entity(EntityType):
            attr = String(constraints=[StaticVocabularyConstraint(['ab', 'abc']),
                                       SizeConstraint(2)])
                          # "auto-fixed" when using:
                          #vocabulary=['ab', 'abc'], maxsize=1)

        with self.assertRaises(BadSchemaDefinition) as cm:
            schema = build_schema_from_namespace(locals().items())
        self.assertEqual("size constraint set to 2 but vocabulary contains string of greater size",
                         str(cm.exception))


    def test_bad_rtype_rdef_conflict(self):
        class foo(RelationType):
            __permissions__={'read': ()}
        class Entity(EntityType):
            foo = String(__permissions__={'add': ()})

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(locals().items())
        self.assertEqual("conflicting values {'add': ()}/{'read': ()} for property __permissions__ of relation 'foo'",
                         str(cm.exception))


    def test_schema(self):
        class Anentity(EntityType):
            rel = SubjectRelation('Anentity', inlined=True)
        class Anotherentity(EntityType):
            rel = SubjectRelation('Anentity')
        class rel(RelationType):
            composite = 'subject'
            cardinality = '1*'
            symmetric = True

        schema = build_schema_from_namespace(locals().items())
        self.assertEqual('<builtin>', schema['Anentity'].package)
        rel = schema['rel']
        self.assertEqual(True, rel.symmetric)
        self.assertEqual(True, rel.inlined)
        self.assertEqual('<builtin>', rel.package)
        rdef1 = rel.rdef('Anentity', 'Anentity')
        self.assertEqual('subject', rdef1.composite)
        self.assertEqual('1*', rdef1.cardinality)
        self.assertEqual('<builtin>', rdef1.package)
        rdef2 = rel.rdef('Anotherentity', 'Anentity')
        self.assertEqual('subject', rdef2.composite)
        self.assertEqual('1*', rdef2.cardinality)
        self.assertEqual('<builtin>', rdef2.package)

    def test_imports(self):
        schema = SchemaLoader().load([self.datadir, self.datadir+'2'], 'Test')
        self.assertEqual({'read': (),
                          'add': (),
                          'update': (),
                          'delete': ()},
                         schema['Affaire'].permissions)
        self.assertEqual([str(r) for r,at in schema['MyNote'].attribute_definitions()],
                          ['date', 'type', 'para', 'text'])

    def test_duplicated_rtype(self):
        loader = SchemaLoader()
        loader.defined = {}
        class RT1(RelationType):
            pass
        loader.add_definition(RT1)
        with self.assertRaises(BadSchemaDefinition) as cm:
            loader.add_definition(RT1)
        self.assertEqual(str(cm.exception), 'duplicated relation type for RT1')

    def test_rtype_priority(self):
        loader = SchemaLoader()
        loader.defined = {}
        class RT1Def(RelationDefinition):
            name = 'RT1'
            subject = 'Whatever'
            object = 'Whatever'
        class RT1(RelationType):
            pass
        loader.add_definition(RT1Def)
        loader.add_definition(RT1)
        self.assertEqual(loader.defined['RT1'], RT1)

    def test_unfinalized_manipulation(self):
        class MyEntity(EntityType):
            base_arg_b = String()
            base_arg_a = Boolean()
            base_sub = SubjectRelation('MyOtherEntity')
        class base_obj(RelationDefinition):
            subject = 'MyOtherEntity'
            object = 'MyEntity'
        class MyOtherEntity(EntityType):
            base_o_obj = SubjectRelation('MyEntity')
        class base_o_sub(RelationDefinition):
            subject = 'MyEntity'
            object = 'MyOtherEntity'
        MyEntity.add_relation(Date(), name='new_arg_a')
        MyEntity.add_relation(Int(), name='new_arg_b')
        MyEntity.add_relation(SubjectRelation('MyOtherEntity'), name="new_sub")
        MyOtherEntity.add_relation(SubjectRelation('MyEntity'), name="new_o_obj")
        class new_obj(RelationDefinition):
            subject = 'MyOtherEntity'
            object = 'MyEntity'
        class new_o_sub(RelationDefinition):
            subject = 'MyEntity'
            object = 'MyOtherEntity'

        schema = build_schema_from_namespace(locals().items())
        self.assertIn('MyEntity', schema.entities())
        my_entity = schema['MyEntity']
        attributes_def = my_entity.attribute_definitions()
        attributes = sorted(attr[0].type for attr in attributes_def)
        self.assertEqual(['base_arg_a', 'base_arg_b', 'new_arg_a', 'new_arg_b'],
                         attributes)
        relations_def = my_entity.relation_definitions()
        relations = sorted( rel[0].type for rel in relations_def)
        self.assertEqual(['base_o_obj', 'base_o_sub', 'base_obj',
                          'base_sub', 'new_o_obj', 'new_o_sub', 'new_obj',
                          'new_sub'],
                         relations)

    def test_post_build_callback(self):
        SchemaLoader.main_schema_directory = 'schema_post_build_callback'
        schema = SchemaLoader().load([self.datadir], 'Test')
        self.assertIn('Toto', schema.entities())


class BuildSchemaTC(TestCase):

    def test_build_schema(self):
        class Question(EntityType):
            number = Int()
            text = String()

        class Form(EntityType):
            title = String()

        class in_form(RelationDefinition):
            subject = 'Question'
            object = 'Form'
            cardinality = '*1'

        schema = build_schema_from_namespace(vars().items())
        entities = [x for x in schema.entities() if not x.final]
        self.assertItemsEqual(['Form', 'Question'], entities)
        relations = [x for x in schema.relations() if not x.final]
        self.assertItemsEqual(['in_form'], relations)

if __name__ == '__main__':
    unittest_main()





# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""unit tests for module yams.schema2sql
"""

import os.path as osp
from cStringIO import StringIO

from logilab.common.testlib import TestCase, unittest_main
from logilab.database import get_db_helper

from yams.reader import SchemaLoader
from yams import schema2sql

schema2sql.SET_DEFAULT = True

DATADIR = osp.abspath(osp.join(osp.dirname(__file__), 'data'))

schema = SchemaLoader().load([DATADIR])


EXPECTED_DATA_NO_DROP = """
CREATE TABLE Affaire(
 sujet varchar(128),
 ref varchar(12),
 inline_rel integer
);
CREATE INDEX affaire_inline_rel_idx ON Affaire(inline_rel);

CREATE TABLE Company(
 name text
);

CREATE TABLE Datetest(
 dt1 timestamp,
 dt2 timestamp,
 d1 date,
 d2 date,
 t1 time,
 t2 time
);

CREATE TABLE Division(
 name text
);

CREATE TABLE EPermission(
 name varchar(100) NOT NULL
);
CREATE INDEX epermission_name_idx ON EPermission(name);

CREATE TABLE Eetype(
 name varchar(64) UNIQUE NOT NULL,
 description text,
 meta boolean,
 final boolean,
 initial_state integer
);
CREATE INDEX eetype_name_idx ON Eetype(name);
CREATE INDEX eetype_initial_state_idx ON Eetype(initial_state);

CREATE TABLE Employee(
);

CREATE TABLE Note(
 date varchar(10),
 type varchar(1),
 para varchar(512)
);

CREATE TABLE Person(
 nom varchar(64) NOT NULL,
 prenom varchar(64),
 sexe varchar(1) DEFAULT 'M',
 promo varchar(6),
 titre varchar(128),
 adel varchar(128),
 ass varchar(128),
 web varchar(128),
 tel integer,
 fax integer,
 datenaiss date,
 test boolean,
 salary float
);
CREATE UNIQUE INDEX unique_e6c2d219772dbf1715597f7d9a6b3892 ON Person(nom,prenom);

CREATE TABLE Salaried(
 nom varchar(64) NOT NULL,
 prenom varchar(64),
 sexe varchar(1) DEFAULT 'M',
 promo varchar(6),
 titre varchar(128),
 adel varchar(128),
 ass varchar(128),
 web varchar(128),
 tel integer,
 fax integer,
 datenaiss date,
 test boolean,
 salary float
);
CREATE UNIQUE INDEX unique_98da0f9de8588baa8966f0b1a6f850a3 ON Salaried(nom,prenom);

CREATE TABLE Societe(
 nom varchar(64),
 web varchar(128),
 tel integer,
 fax integer,
 rncs varchar(32),
 ad1 varchar(128),
 ad2 varchar(128),
 ad3 varchar(128),
 cp varchar(12),
 ville varchar(32)
);

CREATE TABLE State(
 eid integer PRIMARY KEY,
 name varchar(256) NOT NULL,
 description text
);
CREATE INDEX state_name_idx ON State(name);

CREATE TABLE Subcompany(
 name text
);

CREATE TABLE Subdivision(
 name text
);

CREATE TABLE pkginfo(
 modname varchar(30) NOT NULL,
 version varchar(10) DEFAULT '0.1' NOT NULL,
 copyright text NOT NULL,
 license varchar(3),
 short_desc varchar(80) NOT NULL,
 long_desc text NOT NULL,
 author varchar(100) NOT NULL,
 author_email varchar(100) NOT NULL,
 mailinglist varchar(100),
 debian_handler varchar(6)
);


CREATE TABLE concerne_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT concerne_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX concerne_relation_from_idx ON concerne_relation(eid_from);
CREATE INDEX concerne_relation_to_idx ON concerne_relation(eid_to);

CREATE TABLE division_of_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT division_of_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX division_of_relation_from_idx ON division_of_relation(eid_from);
CREATE INDEX division_of_relation_to_idx ON division_of_relation(eid_to);

CREATE TABLE evaluee_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT evaluee_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX evaluee_relation_from_idx ON evaluee_relation(eid_from);
CREATE INDEX evaluee_relation_to_idx ON evaluee_relation(eid_to);

CREATE TABLE next_state_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT next_state_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX next_state_relation_from_idx ON next_state_relation(eid_from);
CREATE INDEX next_state_relation_to_idx ON next_state_relation(eid_to);

CREATE TABLE obj_wildcard_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT obj_wildcard_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX obj_wildcard_relation_from_idx ON obj_wildcard_relation(eid_from);
CREATE INDEX obj_wildcard_relation_to_idx ON obj_wildcard_relation(eid_to);

CREATE TABLE require_permission_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT require_permission_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX require_permission_relation_from_idx ON require_permission_relation(eid_from);
CREATE INDEX require_permission_relation_to_idx ON require_permission_relation(eid_to);

CREATE TABLE state_of_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT state_of_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX state_of_relation_from_idx ON state_of_relation(eid_from);
CREATE INDEX state_of_relation_to_idx ON state_of_relation(eid_to);

CREATE TABLE subcompany_of_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT subcompany_of_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX subcompany_of_relation_from_idx ON subcompany_of_relation(eid_from);
CREATE INDEX subcompany_of_relation_to_idx ON subcompany_of_relation(eid_to);

CREATE TABLE subdivision_of_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT subdivision_of_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX subdivision_of_relation_from_idx ON subdivision_of_relation(eid_from);
CREATE INDEX subdivision_of_relation_to_idx ON subdivision_of_relation(eid_to);

CREATE TABLE subj_wildcard_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT subj_wildcard_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX subj_wildcard_relation_from_idx ON subj_wildcard_relation(eid_from);
CREATE INDEX subj_wildcard_relation_to_idx ON subj_wildcard_relation(eid_to);

CREATE TABLE sym_rel_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT sym_rel_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX sym_rel_relation_from_idx ON sym_rel_relation(eid_from);
CREATE INDEX sym_rel_relation_to_idx ON sym_rel_relation(eid_to);

CREATE TABLE travaille_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT travaille_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX travaille_relation_from_idx ON travaille_relation(eid_from);
CREATE INDEX travaille_relation_to_idx ON travaille_relation(eid_to);

CREATE TABLE works_for_relation (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT works_for_relation_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX works_for_relation_from_idx ON works_for_relation(eid_from);
CREATE INDEX works_for_relation_to_idx ON works_for_relation(eid_to);
"""

class SQLSchemaTC(TestCase):

    def test_known_values(self):
        dbhelper = get_db_helper('postgres')
        output = schema2sql.schema2sql(dbhelper, schema)
        self.assertMultiLineEqual(EXPECTED_DATA_NO_DROP.strip(), output.strip())


if __name__ == '__main__':
    unittest_main()



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Yams packaging information."""
__docformat__ = "restructuredtext en"

# pylint: disable-msg=W0622

# package name
modname = 'yams'

# release version
numversion = (0, 39, 1)
version = '.'.join(str(num) for num in numversion)

# license and copyright
license = 'LGPL'

# short and long description
short_desc = "entity / relation schema"
long_desc = """Yet Another Magic Schema !
A simple/generic but powerful entities / relations schema, suitable
to represent RDF like data. The schema is readable/writable from/to
various formats.
"""

# author name and email
author = "Logilab"
author_email = "devel@logilab.fr"

# home page
web = "http://www.logilab.org/project/%s" % modname

# mailing list
mailinglist = 'mailto://python-projects@lists.logilab.org'

# download place
ftp = "ftp://ftp.logilab.org/pub/%s" % modname

# is there some directories to include with the source installation
include_dirs = []

# executable

scripts = ['bin/yams-check', 'bin/yams-view']

install_requires = [
    'logilab-common >= 0.55.2',
    'logilab-database >= 1.11',
    ]
